@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2006-2010, Benjamin Kaufmann
+// Copyright (c) 2009-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -21,10 +21,12 @@
 #include "alarm.h" // for SIGALRM
 #include <clasp/clasp_facade.h>
 #include <clasp/minimize_constraint.h>
+#include <clasp/satelite.h>
 #include <stdio.h>
 #include <stdarg.h>
 #include <string.h>
 #include <numeric>
+#include <climits>
 
 namespace Clasp { namespace {
 inline std::string prettify(const std::string& str) {
@@ -52,10 +54,10 @@ OutputFormat::RunSummary::Result OutputFormat::RunSummary::result() const {
 	Result r  = result_unknown;
 	if (m != 0) { // SAT
 		if (!e->minimize()) { r = result_sat; }
-		else if (complete)  { r = result_optimum; }
+		else if (complete()){ r = result_optimum; }
 		else                { r = result_sat_opt; }
 	}
-	else if (complete)    { r = result_unsat; }
+	else if (complete()) { r = result_unsat; }
 	return r;
 }
 
@@ -63,12 +65,13 @@ uint64 OutputFormat::RunSummary::models() const {
 	return (ctx.enumerator()->enumerated == 0 || !(ctx.enumerator()->optimize()||consequences)) ? ctx.enumerator()->enumerated : 1;
 }
 
-OutputFormat::OutputFormat(int verbosity) : curr_(0), next_(0), lpStats_(0), verbosity_(verbosity), quiet_(0) {
+OutputFormat::OutputFormat(uint32 verbosity) : curr_(0), next_(0), lpStats_(0), verbosity_(0), quiet_(0) {
 	result[RunSummary::result_unknown] = "UNKNOWN";
 	result[RunSummary::result_unsat]   = "UNSATISFIABLE";
 	result[RunSummary::result_sat]     = "SATISFIABLE";
 	result[RunSummary::result_optimum] = "OPTIMUM FOUND";
 	result[RunSummary::result_sat_opt] = "SATISFIABLE";
+	verbosity_ = (int)std::min(verbosity, (uint32)INT_MAX);
 }
 OutputFormat::~OutputFormat() {
 	delete lpStats_;
@@ -146,16 +149,17 @@ OutputFormat::Model OutputFormat::storeModel(const Assignment& a) {
 /////////////////////////////////////////////////////////////////////////////////////////
 #define printKey(k)                  printf("%s%-*s: ", format_[cat_comment], w_, (k))
 #define printKeyValue(k, fmt, value) printf("%s%-*s: "fmt, format_[cat_comment], w_, (k), (value))
+const char* const lnSep = "----------------------------------------------------------------------------";
 
-DefaultOutput::DefaultOutput(int v, const std::pair<int,int>& q, Format f, char ifs)
+DefaultOutput::DefaultOutput(uint32 v, const std::pair<uint32,uint32>& q, Format f, char ifs)
 	: OutputFormat(v)
 	, prepro_(0)
 	, as_(0)
 	, eom_(0)
 	, header_(1)
+	, ev_(2)
 	, IFS_(ifs) {
-
-	configureFormat(f, q, v, ifs);
+	configureFormat(f, q, ifs);
 }
 
 DefaultOutput::~DefaultOutput() {
@@ -164,7 +168,7 @@ DefaultOutput::~DefaultOutput() {
 }
 
 // sets up format strings and configures verbosity and quiet levels
-void DefaultOutput::configureFormat(Format f, std::pair<int, int> q, int v, char ifs) {
+void DefaultOutput::configureFormat(Format f, PrintPair q, char ifs) {
 	format_[cat_comment]   = "";
 	format_[cat_value]     = "";
 	format_[cat_objective] = "Optimization: ";
@@ -174,11 +178,14 @@ void DefaultOutput::configureFormat(Format f, std::pair<int, int> q, int v, char
 	format_[cat_atom_post] = "";
 	if (f == format_aspcomp) {
 		format_[cat_atom_post] = ".";
+		format_[cat_objective] = "OPT: ";
 		result[RunSummary::result_sat]     = "ANSWER SET FOUND";
 		result[RunSummary::result_unsat]   = "INCONSISTENT";
 		result[RunSummary::result_sat_opt] = "";
-		v        = -1;
-		q.second = print_no;
+		setVerbosity(-1);
+		if (q.second == uint32(print_f_default)) {
+			q.second = print_no;
+		}
 	}
 	else if (f == format_sat09 || f == format_pb09) {
 		format_[cat_comment]   = "c ";
@@ -189,12 +196,11 @@ void DefaultOutput::configureFormat(Format f, std::pair<int, int> q, int v, char
 		if (f == format_pb09) {
 			format_[cat_value_term]= "";
 			format_[cat_atom_pre]  = "x";
-			if (q.first == -1) {
-				q.first = 1;
+			if (q.first == uint32(print_f_default)) {
+				q.first = print_best;
 			}
 		}
 	}
-	setVerbosity(v);
 	setQuiet(q);
 	free((void*)as_);
 	free((void*)eom_);
@@ -233,28 +239,57 @@ void DefaultOutput::comment(int v, const char* fmt, ...) const {
 	}
 }	
 
-void DefaultOutput::reportProgress(Enumerator::ProgressType t, const Solver& s, uint64 maxCfl, uint32 maxL) {
-	const char p[3] = {'R', 'D', 'M'};
-	uint32 fixed    = s.decisionLevel() > 0 ? s.levelStart(1) : s.numAssignedVars();
-	if (verbosity() >= 3 && --header_ == 0) {
-		printf("%s----------------------------------------------------------------------------\n"
-		       "%s[ID:T]     Vars           Constraints         State            Limits\n"
-					 "%s       #free/#fixed   #problem/#learnt  #conflicts/ratio #conflict/#learnt\n"
-					 "%s----------------------------------------------------------------------------\n", 
-					 format_[cat_comment], format_[cat_comment], format_[cat_comment], format_[cat_comment]);
-		header_ = 20;
+void DefaultOutput::reportProgress(const PreprocessEvent& e) {
+	if (e.type == (int)SatElite::SatElite::event_type && verbosity() >= 2) {
+		const SatElite::SatElite::Event& ev = static_cast<const SatElite::SatElite::Event&>(e);
+		printf("\r");
+		if (ev.op != '*' || ev.cur != ev.max) { comment(2, "Preprocessing: %c: %8u/%-8u", ev.op, ev.cur, ev.max); }
+		else                                  { comment(2, "Preprocessing: "); prepro_ = ev.self; }
 	}
-	comment(3, "[%u:%c]|%7u/%-7u|%8u/%-8u|%10"PRIu64"/%-6.3f|%8"PRId64"/%-8"PRId64"|\n"
-		, s.id()
-		, p[t]
-		, s.numFreeVars()
-		, fixed
-		, s.numConstraints()
-		, s.numLearntConstraints()
-		, s.stats.conflicts
-		, s.stats.conflicts/std::max(1.0,double(s.stats.choices))
-		, maxCfl <= (UINT32_MAX) ? (int64)maxCfl:-1
-		, maxL   != (UINT32_MAX) ? (int64)maxL:-1);
+}
+
+void DefaultOutput::reportProgress(const SolveEvent& e) {
+	if (verbosity() < 2 || verbosity() < e.type) { return; }
+	if (e.type != SolveEvent_t::progress_path) {
+		if (ev_ != 1) { ev_ = 1; printf("%s%s\n", format_[cat_comment], lnSep); }
+		if (e.type == SolveEvent_t::progress_state) {
+			const SolveStateEvent& ev = static_cast<const SolveStateEvent&>(e);
+			if (ev.time < 0) { printf("%s[%u:S]| %-15s %-20s  %30c|\n", format_[cat_comment], e.solver->id(), "ENTERING", ev.state, ' '); }
+			else             { printf("%s[%u:S]| %-15s %-20s after %-6.3fs %17c|\n", format_[cat_comment], e.solver->id(), "LEAVING", ev.state, ev.time, ' '); }
+		}
+		else if (e.type == SolveEvent_t::progress_msg) { 
+			printf("%s[%u:G]| %-30s%38c|\n", format_[cat_comment], e.solver->id(), static_cast<const SolveMsgEvent&>(e).msg, ' '); 
+		}
+	}
+	else {
+		const SolvePathEvent& ev = static_cast<const SolvePathEvent&>(e);
+		const Solver& s = *ev.solver;
+		uint32 fixed    = s.decisionLevel() > 0 ? s.levelStart(1) : s.numAssignedVars();
+		if (--header_ == 0) {
+			ev_     = 2;
+			header_ = 20;
+			printf("%s%s\n"
+				"%s[ID:T]     Vars           Constraints         State            Limits\n"
+				"%s       #free/#fixed   #problem/#learnt  #conflicts/ratio #conflict/#learnt\n"
+				"%s%s\n", 
+				format_[cat_comment], lnSep, format_[cat_comment], format_[cat_comment], format_[cat_comment], lnSep);
+		}
+		else if (++ev_ == 2) { printf("%s%s\n", format_[cat_comment], lnSep); }
+		printf("%s[%u:%c]|%7u/%-7u|%8u/%-8u|%10"PRIu64"/%-6.3f|%8"PRId64"/%-8"PRId64"|\n"
+			, format_[cat_comment]
+			, s.id()
+			, static_cast<char>(ev.evType)
+			, s.numFreeVars()
+			, fixed
+			, s.numConstraints()
+			, s.numLearntConstraints()
+			, s.stats.conflicts
+			, s.stats.conflicts/std::max(1.0,double(s.stats.choices))
+			, ev.cLimit <= (UINT32_MAX) ? (int64)ev.cLimit:-1
+			, ev.lLimit != (UINT32_MAX) ? (int64)ev.lLimit:-1
+		);
+	}
+	fflush(stdout);
 }
 
 void DefaultOutput::reportState(int state, bool enter, double time) {
@@ -270,30 +305,20 @@ void DefaultOutput::reportState(int state, bool enter, double time) {
 			comment(2, "Preprocessing: ");
 		}
 		else if (state == ClaspFacade::state_solve) {
-			comment(1, "Solving...\n");
+			if (verbosity() > 1) { printf("%s%s\n%sSolving...%66c\n", format_[cat_comment], lnSep, format_[cat_comment], '|'); }
+			else                 { comment(1, "Solving...\n"); }
 		}
 	}
-	else if (!enter && verbosity() > 1 && (state == ClaspFacade::state_read || state == ClaspFacade::state_preprocess)) {
-		printf("%.3f", time);
-		if (prepro_ != 0) {
-			printf(" (ClRemoved: %u ClAdded: %u LitsStr: %u)", prepro_->stats.clRemoved, prepro_->stats.clAdded, prepro_->stats.litsRemoved);
+	else if (!enter && verbosity() > 1) {
+		if (state == ClaspFacade::state_read || state == ClaspFacade::state_preprocess) {
+			printf("%.3f", time);
+			if (prepro_ != 0) { printf(" (ClRemoved: %u ClAdded: %u LitsStr: %u)", prepro_->stats.clRemoved, prepro_->stats.clAdded, prepro_->stats.litsRemoved); }
+			printf("\n");
 		}
-		printf("\n"); fflush(stdout);
+		else if (state == ClaspFacade::state_solve) { printf("%s%s\n", format_[cat_comment], lnSep); } 
 	}
 }	
 
-void DefaultOutput::reportPreProgress(char type, const SatPreprocessor& p, uint32 min, uint32 max) {
-	printf("\r");
-	if (type != '*' || min != max) {
-		comment(2, "Preprocessing: %c: %8u/%-8u", type, min, max);
-	}
-	else {
-		comment(2, "Preprocessing: ");
-		prepro_ = &p;
-	}
-	
-}
-
 void DefaultOutput::printModel(const Model& m, const SymbolTable& index, const Enumerator& en) {
 	comment(1, "Answer: %"PRIu64"\n", en.enumerated);
 	printf("%s", format_[cat_value]);
@@ -359,10 +384,10 @@ void DefaultOutput::printResult(const RunSummary& sol, const SolveStats** stats,
 	printf("%s%s\n", format_[cat_result], result[sol.result()]);
 	if (verbosity() > (0 - (num!=0))) {
 		printf("\n");
-		if      (sol.term == SIGALRM) { printKeyValue("TIME LIMIT", "%d\n", 1);  }
-		else if (sol.term != 0)       { printKeyValue("INTERRUPTED", "%d\n", 1); }
+		if      (sol.termSig() == SIGALRM) { printKeyValue("TIME LIMIT", "%d\n", 1);  }
+		else if (sol.termSig() != 0)       { printKeyValue("INTERRUPTED", "%d\n", 1); }
 		printKey("Models");
-		if (!sol.complete) {
+		if (!sol.complete()) {
 			char buf[64];
 			int wr    = sprintf(buf, "%"PRIu64, sol.models());
 			buf[wr]   = '+';
@@ -377,10 +402,10 @@ void DefaultOutput::printResult(const RunSummary& sol, const SolveStats** stats,
 				printKeyValue("  Enumerated", "%"PRIu64"\n", en.enumerated);
 			}
 			if (sol.consequences) {
-				printf("%s  %-*s: %s\n", format_[cat_comment], w_-2, sol.consequences, sol.complete?"yes":"unknown");
+				printf("%s  %-*s: %s\n", format_[cat_comment], w_-2, sol.consequences, sol.complete()?"yes":"unknown");
 			}
 			if (en.minimize()) {
-				printKeyValue("  Optimum", "%s\n", sol.complete?"yes":"unknown");
+				printKeyValue("  Optimum", "%s\n", sol.complete()?"yes":"unknown");
 				if (en.optimize()) {
 					printKey("Optimization");
 					printOptimizeValues(*en.minimize());
@@ -413,7 +438,7 @@ void DefaultOutput::printStats(const RunSummary& sol, const SolveStats** stats,
 	printProblemStats(pr);
 	printLemmaStats(*st);
 	printJumpStats(*st);
-	printThreadStats(stats, num, sol.winner);
+	printThreadStats(stats, num, sol);
 	fflush(stdout);
 }
 
@@ -422,7 +447,7 @@ void DefaultOutput::printSolveStats(const SolveStats& st) const {
 	printKeyValue("Conflicts", "%"PRIu64"\n", st.conflicts);
 	printKeyValue("Restarts", "%-6"PRIu64"", st.restarts);
 	if (st.restarts) {
-		printf(" (Average: %.2f Last: %"PRIu64")", average(st.analyzed, st.restarts), st.analyzed - st.lRestart);
+		printf(" (Average: %.2f Last: %"PRIu64")", average(st.analyzed, st.restarts), st.cflLast);
 	}
 	printf("\n");
 }
@@ -481,11 +506,7 @@ void DefaultOutput::printLemmaStats(const SolveStats& st) const {
 	printKeyValue("  Loop", "%-6"PRIu64, st.learnts[1]);
 	printf(" (Average Length: %.1f) \n", average(st.lits[1], st.learnts[1]));
 	printKeyValue("  Other", "%-6"PRIu64, st.learnts[2]);
-	printf(" (Average Length: %.1f"     , average(st.lits[2], st.learnts[2]));
-	if (st.parallel != 0) {  
-		printf(" Unit: %"PRIu64" Average Jumps: %.2f", st.parallel->imps, average(st.parallel->jumps, st.parallel->imps));
-	}
-	printf(")\n");
+	printf(" (Average Length: %.1f)\n"  , average(st.lits[2], st.learnts[2]));
 	printKeyValue("  Deleted", "%-6"PRIu64"\n", st.deleted);
 }
 
@@ -505,16 +526,17 @@ void DefaultOutput::printJumpStats(const SolveStats& st) const {
 		printKeyValue("Average Jump Length", "%-6.1f", average(js.jumpSum, st.analyzed));
 		printf(" (Executed: %.1f)\n", average(js.jumpSum-js.boundSum, st.analyzed));
 		printKeyValue("Average Bound Length", "%-6.1f\n", average(js.boundSum,js.bJumps));
-		printKeyValue("Average Model Length", "%-6.1f\n", average(st.modLits,st.models));
+		printKeyValue("Average Model Length", "%-6.1f\n", average(js.modLits,st.models));
 		w_ -= 8;
 	}
 }
 
-void DefaultOutput::printThreadStats(const SolveStats** stats, std::size_t num, uint32 winner) const {
+void DefaultOutput::printThreadStats(const SolveStats** stats, std::size_t num, const RunSummary& sol) const {
 	if (num > 1 || stats[0]->parallel) {
 		printf("\n");
 		comment(0, "============ PARALLEL STATISTICS ============\n\n");
-		printKeyValue("Winner",  "%u\n", winner);
+		printKeyValue("Threads", "%u\n", sol.ctx.numSolvers());
+		printKeyValue("Winner",  "%d\n", sol.winner());
 		DefaultOutput::printParallelStats(*stats[0], true);
 		for (std::size_t i = 1; i < num; ++i) {
 			comment(0, "[Thread %u]\n", (uint32)i-1);
@@ -530,20 +552,20 @@ void DefaultOutput::printThreadStats(const SolveStats** stats, std::size_t num,
 void DefaultOutput::printParallelStats(const SolveStats& stats, bool accu) const {
 	if (stats.parallel) {
 		const ParallelStats& ps = *stats.parallel;
-		uint64 solved = ps.gps ? ps.gps-uint64(ps.term) : 0;
+		uint64 sumLem = stats.learnts[0] + stats.learnts[1];
 		if (!accu) {
 			printKeyValue("CPU Time",  "%.3fs\n", ps.cpuTime);
 			printKeyValue("Models", "%"PRIu64"\n", stats.models);
 		}
-		printKeyValue("Distributed", "%"PRIu64"\n", ps.shared);
-		printKeyValue("Splits", "%"PRIu64"\n", ps.splits);
-		printKeyValue("Problems", "%-6"PRIu64, ps.gps);
-		if (!accu) {
-			printf(" (Solved: %"PRIu64" Average Length: %.2f) \n", solved, average(ps.gpLits, ps.gps));
-		}
-		else {
-			printf(" (Average Length: %.2f) \n\n", average(ps.gpLits, ps.gps));
-		}
+		printKeyValue("Distributed", "%-6"PRIu64, ps.distributed);
+		printf(" (Ratio: %.2f Average LBD: %.2f) \n", ps.distributed/std::max(1.0, double(sumLem)), average(ps.sumLbd, ps.distributed));
+		printKeyValue("Integrated", "%-6"PRIu64, ps.integrated);
+		if (accu) { printf(" (Ratio: %.2f ", ps.integrated/std::max(1.0, double(ps.distributed))); }
+		else      { printf(" ("); }
+		printf("Unit: %"PRIu64" Average Jumps: %.2f)\n", ps.imps, average(ps.jumps, ps.imps));
+		printKeyValue("Splits", "%"PRIu64"\n", (uint64)ps.splits);
+		printKeyValue("Problems", "%-6"PRIu64, (uint64)ps.gps);
+		printf(" (Average Length: %.2f) \n\n", average(ps.gpLits, ps.gps));
 	}
 }
 
@@ -553,7 +575,7 @@ void DefaultOutput::printParallelStats(const SolveStats& stats, bool accu) const
 #undef printKeyValue
 #undef printKey
 
-JsonOutput::JsonOutput(int v, const std::pair<int,int>& q) : OutputFormat(v), indent_(0), open_(""), hasModel_(false), hasWitness_(false) {
+JsonOutput::JsonOutput(uint32 v, const std::pair<uint32,uint32>& q) : OutputFormat(v), indent_(0), open_(""), hasModel_(false), hasWitness_(false) {
 	setQuiet(q);
 }
 
@@ -611,8 +633,6 @@ void JsonOutput::endObject(ObjType t) {
 	open_ = ",\n";
 }
 
-void JsonOutput::reportPreProgress(char, const SatPreprocessor&, uint32, uint32) { }
-
 void JsonOutput::reportState(int state, bool enter, double) {
 	if (indent_ == 0) {
 		open_ = "";
@@ -701,19 +721,19 @@ void JsonOutput::printResult(const RunSummary& sol, const SolveStats** stats, st
 	printKeyValue("Result", result[sol.result()]);
 	const Enumerator& en = *sol.ctx.enumerator();
 	startObject("Stats");
-	if      (sol.term == SIGALRM) { printKeyValue("TIME LIMIT", uint32(1));  }
-	else if (sol.term != 0)       { printKeyValue("INTERRUPTED", uint32(1)); }
+	if      (sol.termSig() == SIGALRM) { printKeyValue("TIME LIMIT", uint32(1));  }
+	else if (sol.termSig() != 0)       { printKeyValue("INTERRUPTED", uint32(1)); }
 	printKeyValue("Models", sol.models());
-	printKeyValue("Complete", sol.complete ? "yes" : "no");
+	printKeyValue("Complete", sol.complete() ? "yes" : "no");
 	if (en.enumerated) {
 		if (sol.models() != en.enumerated) {
 			printKeyValue("Enumerated", en.enumerated);
 		}
 		if (sol.consequences) {
-			printKeyValue(sol.consequences, sol.complete ? "yes":"unknown");
+			printKeyValue(sol.consequences, sol.complete() ? "yes":"unknown");
 		}
 		if (en.minimize()) {
-			printKeyValue("Optimum", sol.complete?"yes":"unknown");
+			printKeyValue("Optimum", sol.complete()?"yes":"unknown");
 			if (en.optimize()) {
 				printOptimize(*en.minimize());
 			}
@@ -740,14 +760,13 @@ void JsonOutput::printResult(const RunSummary& sol, const SolveStats** stats, st
 		printKeyValue("Sum", sum);
 		printKeyValue("Binary", p.constraints_binary);
 		printKeyValue("Ternary", p.constraints_ternary);
-		printKeyValue("Other", p.constraints_ternary);
 		endObject(); // Constraints
 		if (lpStats()) {
 			printLpStats(*lpStats());
 		}
 		printLemmaStats(*st);
 		printJumpStats(*st);
-		printThreadStats(stats, num, sol.winner);
+		printThreadStats(stats, num, sol);
 		endObject(); // Stats
 	}
 	else {
@@ -813,7 +832,6 @@ void JsonOutput::printLemmaStats(const SolveStats& st) {
 	printKeyValue("Sum", sum);
 	printKeyValue("Binary", st.binary);
 	printKeyValue("Ternary", st.ternary);
-	printKeyValue("Other", sum - st.binary - st.ternary);
 	printKeyValue("Conflict", st.learnts[0]);
 	printKeyValue("Loop", st.learnts[1]);
 	printKeyValue("Other", st.learnts[2]);
@@ -839,17 +857,17 @@ void JsonOutput::printJumpStats(const SolveStats& st) {
 		printKeyValue("AvgJump", average(js.jumpSum, st.analyzed));
 		printKeyValue("AvgJumpEx", average(js.jumpSum-js.boundSum, st.analyzed));
 		printKeyValue("AvgBound", average(js.boundSum,js.bJumps));
-		printKeyValue("AvgModel", average(st.modLits,st.models));
+		printKeyValue("AvgModel", average(js.modLits,st.models));
 		endObject();
 	}
 }
 
-void JsonOutput::printThreadStats(const SolveStats** stats, std::size_t num, uint32 winner) {
+void JsonOutput::printThreadStats(const SolveStats** stats, std::size_t num, const RunSummary& sol) {
 	if (stats[0]->parallel) {
 		startObject("Parallel");
 		const ParallelStats& ps = *stats[0]->parallel;
-		printKeyValue("Winner", winner);
-		printKeyValue("Distributed", ps.shared);
+		if (sol.winner() >= 0) printKeyValue("Winner", (uint32)sol.winner());
+		printKeyValue("Distributed", ps.distributed);
 		printKeyValue("Splits", ps.splits);
 		printKeyValue("Problems", ps.gps);
 		printKeyValue("AvgGPLength", average(ps.gpLits, ps.gps));
@@ -871,7 +889,6 @@ void JsonOutput::printThreadStats(const SolveStats** stats, std::size_t num, uin
 void JsonOutput::printParallelStats(const SolveStats& stats) {
 	if (stats.parallel) {
 		const ParallelStats& ps = *stats.parallel;
-		uint64 solved = ps.gps ? ps.gps-uint64(ps.term) : 0;
 		startObject("Basic");
 		printKeyValue("CPU Time", ps.cpuTime);
 		printKeyValue("Models", stats.models);
@@ -879,10 +896,9 @@ void JsonOutput::printParallelStats(const SolveStats& stats) {
 		printKeyValue("Conflicts", stats.conflicts);
 		printKeyValue("Restarts", stats.restarts);
 		printKeyValue("Problems", ps.gps);
-		printKeyValue("Solved", solved);
 		printKeyValue("AvgGPLength", average(ps.gpLits, ps.gps));
 		printKeyValue("Splits", ps.splits);
-		printKeyValue("Distributed", ps.shared);
+		printKeyValue("Distributed", ps.distributed);
 		endObject();
 	}
 }
@@ -890,7 +906,7 @@ void JsonOutput::printParallelStats(const SolveStats& stats) {
 }
 
 using namespace Clasp;
-#if defined(WITH_CLASPRE_SUPPORT)
+#if WITH_CLASPRE
 #include <program_opts/program_options.h>
 #include <program_opts/typed_value.h>
 #include <sstream>
@@ -900,13 +916,13 @@ void Options::initOptions(ProgramOptions::OptionContext& root) {
 	OptionGroup pre("Claspre Options");
 	pre.addOptions()
 		("list-features"   , flag(listFeatures),"Print feature names (in --features=C1 format)")
-		("features"   , storeTo(features, &Options::mapFormat),
+		("features!"   , storeTo(features, &Options::mapFormat),
 			"Print features in selected format\n"
-			"      Valid: V, C1, C2, no\n"
+			"      %A: {V|C1|C2|C3}\n"
 			"        V : Print features in human-readable format\n"
 			"        C1: Print in compact format; iterated features in separate lines\n"
 			"        C2: Print in compact format; iterated features all in one line\n"
-			"        C3: Print in C2 format but only if search state is unknown on exit\n")
+			"        C3: Print in C2 format but only if search state is unknown on exit")
 	;
 	root.add(pre);
 }
@@ -926,21 +942,16 @@ void Options::printFeatures() const {
 }
 
 bool Options::mapFormat(const std::string& s, int& format) {
-	if (s == "no")                 return true;
-	if (s.empty() || s.size() > 2) return false;
-	if (s[0] == 'V' || s[0] == 'v') {
-		format = features_verbose;
-		return s.size() == 1;
+	if (s.size() == 1 && s[0] == 'V' || s[0] == 'v') {
+		format = features_verbose; 
+		return true;
 	}
-	else if (s[0] == 'C' || s[0] == 'c') {
-		switch (s[1]) {
-			case '1': format = features_compact_1; break;
-			case '2': format = features_compact_2; break;
-			case '3': format = features_compact_3; break;
-			default: return false;
-		}
+	if (s.size() == 2 && s[0] == 'C' || s[0] == 'c') {
+		if (s[1] == '1') { format = features_compact_1; return true; }
+		if (s[1] == '2') { format = features_compact_2; return true; }
+		if (s[1] == '3') { format = features_compact_3; return true; }
 	}
-	return true;
+	return s == "no" && (format = features_no, true);
 }
 
 /////////////////////////////////////////////////////////////////////////////////////////
@@ -951,7 +962,8 @@ public:
 	ClaspreOutput(Options::FeatureFormat f, int v, const std::pair<int,int>& q, DefaultOutput::Format outf);
 	void initSolve(const Solver& s, ProgramBuilder* api);
 	void printResult(const RunSummary& sol, const SolveStats** st, std::size_t num);
-	void reportProgress(Enumerator::ProgressType t, const Solver& s, uint64 maxCfl, uint32 maxL);
+	void reportProgress(const PreprocessEvent&) {}
+	void reportProgress(const SolveEvent& ev);
 	void reportState(int state, bool enter, double time) {
 		if (format_ ==  Options::features_verbose) {
 			DefaultOutput::reportState(state, enter, time);
@@ -977,9 +989,13 @@ void ClaspreOutput::initSolve(const Solver& s, ProgramBuilder* api) {
 	}
 }
 
-void ClaspreOutput::reportProgress(Enumerator::ProgressType t, const Solver& s, uint64 maxC, uint32 maxL) {
-	if (t != Enumerator::progress_restart) return;
+void ClaspreOutput::reportProgress(const SolveEvent& e) {
+	if (e.type != SolveEvent_t::progress_path || static_cast<const SolvePathEvent&>(e).evType != SolvePathEvent::event_restart) { return; }
+	const SolvePathEvent& ev = static_cast<const SolvePathEvent&>(e);
 	std::stringstream temp;
+	const Solver& s = *ev.solver;
+	uint64 maxC     = ev.cLimit;
+	uint32 maxL     = ev.lLimit;
 	if (format_ == Options::features_verbose) {
 		printf("\n[Iteration %u]\n", starts_);
 		printf("%-22s: %u\n", "maxLearnt", maxL);
@@ -1016,14 +1032,14 @@ void ClaspreOutput::reportProgress(Enumerator::ProgressType t, const Solver& s,
 }
 
 void ClaspreOutput::printResult(const RunSummary& sol, const SolveStats** stats, std::size_t num) {
-	bool hasSolution = sol.complete || sol.models() > 0;
+	bool hasSolution = sol.complete() || sol.models() > 0;
 	if (format_ == Options::features_verbose) {
 		DefaultOutput::printResult(sol, stats, num);
 	}
 	else {
 		if (starts_ == 0) {
 			Solver empty;
-			reportProgress(Enumerator::progress_restart, empty, 0, 0);
+			reportProgress(SolvePathEvent(empty, SolvePathEvent::event_restart, 0, 0));
 		}
 		if (!iterationBuffer_.empty()) {
 			printf("\n%s\n", iterationBuffer_.c_str());
@@ -1059,7 +1075,7 @@ void ClaspreOutput::printResult(const RunSummary& sol, const SolveStats** stats,
 					 << (st.conflicts-st.analyzed) << "," << st.analyzed << "," << js.bJumps << ","
 					 << js.jumpSum << "," << (js.jumpSum - js.boundSum) << "," << percent(js.jumpSum - js.boundSum, js.jumpSum) << ","
 					 << js.maxJump << "," << js.maxJumpEx << "," << js.maxBound << "," << average(js.jumpSum, st.analyzed) << "," << average(js.jumpSum-js.boundSum, st.analyzed) << ","
-					 << average(js.boundSum,js.bJumps) << "," << average(st.modLits,st.models) << ","
+					 << average(js.boundSum,js.bJumps) << "," << average(js.modLits,st.models) << ","
 					 << learntSum << "," << percent(st.binary, learntSum) << "," << percent(st.ternary, learntSum) << "," 
 					 << percent(learntSum-st.binary-st.ternary, learntSum) << ","
 					 << st.learnts[0] << "," << average(st.lits[0], st.learnts[0]) << ","
@@ -1072,11 +1088,11 @@ void ClaspreOutput::printResult(const RunSummary& sol, const SolveStats** stats,
 	}
 }
 
-OutputFormat* Options::createOutput(int v, std::pair<int,int> q, DefaultOutput::Format f) { 
+OutputFormat* Options::createOutput(uint32 v, DefaultOutput::PrintPair q, DefaultOutput::Format f) { 
 	if (features != Options::features_verbose) {
 		v = 0;
 		if (features != Options::features_compact_3) {
-			q.first = q.second = 2;
+			q.first = q.second = DefaultOutput::print_no;
 		}
 	}
 	return new ClaspreOutput(static_cast<Options::FeatureFormat>(features), v, q, f);
@@ -1088,6 +1104,6 @@ namespace Claspre {
 void Options::initOptions(ProgramOptions::OptionContext&) {}
 bool Options::validateOptions(const ProgramOptions::ParsedOptions&) { return true; }
 void Options::printFeatures() const {}
-OutputFormat* Options::createOutput(int, std::pair<int,int>, DefaultOutput::Format) { return 0; }
+OutputFormat* Options::createOutput(uint32, DefaultOutput::PrintPair, DefaultOutput::Format) { return 0; }
 }
 #endif
