@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2006-2010, Benjamin Kaufmann
+// Copyright (c) 2006-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -26,7 +26,7 @@
 #endif
 
 #if !defined(CLASP_VERSION)
-#define CLASP_VERSION "2.0.6"
+#define CLASP_VERSION "2.1.0"
 #endif
 #if !defined(CLASP_LEGAL)
 #define CLASP_LEGAL \
@@ -36,19 +36,27 @@
 "There is NO WARRANTY, to the extent permitted by law."
 #endif
 
-#if defined(WITH_THREADS)
-#undef WITH_THREADS
-#endif
-#ifndef DISABLE_MULTI_THREADING
-#define WITH_THREADS 1
-#else
-#define WITH_THREADS 0
+#if !defined(WITH_THREADS)
+#error Invalid thread configuration - use WITH_THREADS=0 for single-threaded or WITH_THREADS=1 for multi-threaded version of libclasp!
 #endif
 
+#if WITH_THREADS
+#include <clasp/parallel_solve.h>
+typedef Clasp::mt::ParallelSolveOptions SolveOptions;
+#else
+#include <clasp/solve_algorithms.h>
+namespace Clasp {
+struct SolveOptions {
+	SolveLimits   limit;  /**< Solve limit (disabled by default). */
+	void   createSolveObject(SolveAlgorithm*& out, SharedContext&, SolverConfig**) const { out = new SimpleSolve(limit); }
+	static uint32 supportedSolvers()   { return 1; }
+	static uint32 recommendedSolvers() { return 1; }
+};
+}
+#endif
 #include <clasp/literal.h>
 #include <clasp/solver.h>
 #include <clasp/enumerator.h>
-#include <clasp/solve_algorithms.h>
 #include <clasp/heuristics.h>
 #include <clasp/lookahead.h>
 #include <clasp/program_builder.h>
@@ -67,102 +75,39 @@ namespace Clasp {
 
 /////////////////////////////////////////////////////////////////////////////////////////
 // Parameter configuration
-/////////////////////////////////////////////////////////////////////////////////////////
-//! Options that control the decision heuristic.
-struct HeuristicOptions {
-	HeuristicOptions();
-	typedef Lookahead::Type LookaheadType;
-	DecisionHeuristic* createHeuristic() const;
-	std::string   name;        /**< Name of decision heuristic. */
-	LookaheadType lookahead;   /**< Type of lookahead. */
-	int           lookaheadNum;/**< Number of times lookahead should be used. */
-	int           loops;       /**< Consider loops in heuristic (0: no, 1: yes, -1: let heuristic decide). */
-	union Extra {
-	int      berkMax;          /**< Consider at most berkMax learnt constraints (Only for Berkmin). */
-	int      vmtfMtf;          /**< Move up to vmtfMtf conflict literals to the front (Only for Vmtf).    */
-	int      vsidsDecay;       /**< Use 1.0/0.vsidsDecay as decay factor (Only for Vsids). */
-	} extra;
-	bool     berkMoms;         /**< Only for Berkmin. */
-	bool     berkHuang;	       /**< Only for Berkmin. */
-	bool     berkOnce;         /**< Only for Berkmin. */
-	bool     nant;             /**< Only for unit.    */
-};
-
-//! Local options holding options for one solver instance.
-class LocalOptions {
-public:
-	typedef DefaultUnfoundedCheck::ReasonStrategy LoopMode;
-	explicit LocalOptions(Solver* s = 0);
-	~LocalOptions();
-	void setSolver(Solver* s) { solver_ = s; }
-	bool validate(std::string& err);
-	HeuristicOptions& heuristic();
-	void              initFrom(const LocalOptions& other);
-	void              applyHeuristic();
-	Solver&           solver() { return *solver_; }
-	SolveParams       solve;   /**< Strategies used during solving. */ 
-	LoopMode          loopRep; /**< How to represent loops? */
-private:
-	LocalOptions(const LocalOptions&);
-	LocalOptions& operator=(const LocalOptions&);
-	Solver*           solver_;
-	HeuristicOptions* heuristic_;
-};
-
-//! Options for configuring parallel (multi-threaded) solving.
-struct ThreadOptions {  
-	ThreadOptions() : forceGP(false) {}
-	struct Distribution { /**< Nogood distribution options. */
-		Distribution();
-		enum Filter { filter_no = 0, filter_gp = 1, filter_sat = 2, filter_heuristic = 3 };
-		uint32 grace;       /**< Lower bound on number of shared nogoods to keep. */
-		uint8  lbd;         /**< Upper bound on lbd for sharing nogoods. */
-		uint8  types;       /**< Types of nogoods to share. */
-		uint8  filter;      /**< Filter for integrating shared nogoods (one of Filter). */
-		bool   copyProblem; /**< Copy or share initial problem constraints. */
-	} distribute;         /**< Nogood distribution parameters. */
-	struct GRestarts {    /**< Options for configuring global restarts. */
-		GRestarts():maxR(0) {}
-		uint32           maxR;
-		ScheduleStrategy sched;
-	} restarts;             /**< Global restart strategy. */
-	bool        forceGP;    /**< Force guiding path scheme. */
-};
-
-
+/////////////////////////////////////////////////////////////////////////////////////////	
 //! Global options controlling global solving algorithms.
 struct GlobalOptions {
 public:
-	typedef ProgramBuilder::EqOptions        EqOptions;
-	typedef Enumerator::ProgressReport       Progress;
+	typedef ProgramBuilder::EqOptions EqOptions;
 	typedef std::auto_ptr<Enumerator> EnumPtr;
 	enum EnumMode { enum_auto = 0, enum_bt = 1, enum_record = 2, enum_consequences = 4, enum_brave = 5, enum_cautious = 6 };
 	GlobalOptions();
-	Enumerator*     initEnumerator(Enumerator::Report* r = 0);
+	Enumerator*     createEnumerator(Enumerator::Report* r = 0);
 	bool consequences() const { return enumerate.consequences(); }
 	const char* cbType()const { return consequences() ? (enumerate.mode == enum_brave  ? "Brave" : "Cautious") : "none"; }
 	SharedContext ctx;    /**< Context-object used by all solvers. */
-	ThreadOptions thread; /**< Options for parallel solving. */
+	SolveOptions  solve;  /**< Additional options for solving. */
 	EqOptions     eq;     /**< Options for equivalence preprocessing. */
 	struct Optimize {     /**< Optimization options. */
 		Optimize() : hierarch(0), no(false), all(false) {}
-		WeightVec vals;     /**< Initial values for optimize statements. */
-		int    hierarch;    /**< Use hierarchical optimization scheme. */
+		SumVec vals;        /**< Initial values for optimize statements. */
+		uint32 hierarch;    /**< Use hierarchical optimization scheme. */
 		bool   no;          /**< Ignore optimize statements. */
 		bool   all;         /**< Compute all models <= vals. */
-	}    opt;
+	}             opt;
 	struct EnumOptions {  /**< Enumeration options. */
-		EnumOptions() : progress(0), numModels(-1), mode(enum_auto), projectOpts(7), project(false)
-		              , restartOnModel(false), onlyPre(false)  {}
+		EnumOptions() : numModels(-1), mode(enum_auto), projectOpts(7), project(false)
+		              , maxSat(false), restartOnModel(false), onlyPre(false)  {}
 		bool consequences() const { return (mode & enum_consequences) != 0; }
-		Progress* progress;      /**< Enable progress reporting? */
 		int       numModels;     /**< Number of models to compute. */
 		EnumMode  mode;          /**< Enumeration type to use. */
-		int       projectOpts;   /**< Options for projection. */
+		uint32    projectOpts;   /**< Options for projection. */
 		bool      project;       /**< Enable projection. */
+		bool      maxSat;        /**< Treat DIMACS input as MaxSat */
 		bool      restartOnModel;/**< Restart after each model. */
 		bool      onlyPre;       /**< Stop after preprocessing step? */
-	} enumerate;
+	}             enumerate;
 };
 
 class ClaspFacade;
@@ -189,20 +134,32 @@ private:
 //! Parameter-object that groups & validates options.
 class ClaspConfig : public GlobalOptions {
 public:
+	typedef DefaultUnfoundedCheck::ReasonStrategy LoopMode;
+	typedef Lookahead::Type                       LookaheadType;
+	typedef DecisionHeuristic                     Heuristic;
+	enum HeuristicType { heu_none = 0, heu_berkmin = 1, heu_vsids = 2, heu_vmtf = 3, heu_unit = 4 };
+	static Heuristic* createHeuristic(const SolverStrategies& str);
 	ClaspConfig();
 	~ClaspConfig();
-	uint32        numThreads() const;
-	LocalOptions* threadConfig(uint32 i) const;
-	LocalOptions* master() const { return threadConfig(0); }
+	SolverConfig* getSolver(uint32 i) const { return solvers_.at(i);}
+	SolverConfig* master()            const { return getSolver(0); }
+	uint32        numSolvers()        const { return (uint32)solvers_.size(); }
+	SolverConfig* addSolver()               { addSolver(new Solver()); return solvers_.back(); }
+	SolverConfig**solvers()                 { return &solvers_[0]; }
+	void          setMaxSolvers(uint32 i)   { removeSolvers(i); }
+	void          reserveSolvers(uint32 num);
+	uint32        removeSolvers(uint32 id);
 	bool          validate(std::string& err);
+	bool          validate(SolverConfig& sc, std::string& err);
 	void          applyHeuristic();
+	void          applyHeuristic(SolverConfig& sc);
 	void          reset();
-	void          setThreads(uint32 i);
 private:
 	ClaspConfig(const ClaspConfig&);
 	ClaspConfig& operator=(const ClaspConfig&);
-	typedef PodVector<LocalOptions*>::type SolverOptions;
-	SolverOptions solvers;
+	void          addSolver(Solver* s);
+	typedef PodVector<SolverConfig*>::type SolverConfigs;
+	SolverConfigs solvers_;
 };
 /////////////////////////////////////////////////////////////////////////////////////////
 // ClaspFacade
@@ -247,15 +204,6 @@ public:
 		virtual void warning(const char* msg)       = 0;
 	};
 	ClaspFacade();
-	/*!
-	 * Returns the number of hardware threads available on this system
-	 * or 1 if this information is not available.
-	 * 
-	 * \note The number of hardware threads typically depends on the number of CPUs or cores 
-	 * or hyperthreading units. 
-	 */
-	static uint32 hardware_concurrency();
-
 	/*!
 	 * Solves the problem given in problem using the given configuration.
 	 * \pre config is valid, i.e. config.valid() returned true
@@ -264,7 +212,7 @@ public:
 	 * \note If config.onlyPre is true, solve() returns after
 	 * the preprocessing step (i.e. once the solver is prepared) and does not start a search.
 	 */
-	void solve(Input& problem, ClaspConfig& config, Callback* c);
+	void solve(Input& problem, ClaspConfig& config, Callback* c) { return solve(problem, config, 0, c); }
 
 	/*!
 	 * Incrementally solves the problem given in problem using the given configuration.
@@ -284,7 +232,7 @@ public:
 	 * \endcode
 	 * 
 	 */
-	void solveIncremental(Input& problem, ClaspConfig& config, IncrementalControl& inc, Callback* c);
+	void solveIncremental(Input& problem, ClaspConfig& config, IncrementalControl& inc, Callback* c) { solve(problem, config, &inc, c); }
 
 	//! Returns the result of a computation.
 	Result result() const { return result_; }
@@ -297,7 +245,7 @@ public:
 	//! Returns the current input problem.
 	Input* input() const { return input_; }
 	//! Tries to terminate an active search.
-	bool   terminate() const {  return ctrl_ && ctrl_->terminate(); }
+	bool   terminate() const { return ctrl_ && ctrl_->terminate(); }
 	
 	const ClaspConfig* config() const { return config_; }
 
@@ -315,14 +263,16 @@ public:
 private:
 	ClaspFacade(const ClaspFacade&);
 	ClaspFacade& operator=(const ClaspFacade&);
+	void solve(Input& problem, ClaspConfig& config, IncrementalControl* inc, Callback* c);
 	struct AutoState {
 		AutoState(ClaspFacade* f, State s) : self_(f), state_(s) { self_->setState(s, event_state_enter); }
-		~AutoState() { self_->setState(state_, event_state_exit); delete self_->ctrl_; self_->ctrl_ = 0; }
+		~AutoState() { self_->setState(state_, event_state_exit); }
 		ClaspFacade* self_;
 		State        state_;
 	};
 	typedef SingleOwnerPtr<ProgramBuilder> Api;
 	typedef SingleOwnerPtr<SharedDependencyGraph> GraphPtr;
+	typedef SingleOwnerPtr<Enumerator>            EnumPtr;
 	// -------------------------------------------------------------------------------------------  
 	// Status information
 	void setState(State s, Event e)          { state_ = s; if (cb_) cb_->state(e, *this); }
@@ -346,10 +296,10 @@ private:
 	void   init(Input&, ClaspConfig&, IncrementalControl*, Callback* c);
 	bool   read();
 	bool   preprocess();
-	void   setProblemSize() const;
+	bool   solve(const LitVec& assume);
 	bool   initEnumeration(SharedMinimizeData* min);
+	bool   initContextObject(SharedContext& ctx) const;
 	void   setGraph();
-	void   initSolveObject(ClaspConfig& config);
 	// -------------------------------------------------------------------------------------------
 	ClaspConfig*           config_;
 	IncrementalControl*    inc_;
@@ -357,6 +307,7 @@ private:
 	Input*                 input_;
 	SolveAlgorithm*        ctrl_;
 	GraphPtr               graph_;
+	EnumPtr                enum_;
 	Api                    api_;
 	Result                 result_;
 	State                  state_;
