@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2006-2007, Benjamin Kaufmann
+// Copyright (c) 2006-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -54,8 +54,25 @@ struct Constraint_t {
 		learnt_other      = 3, /**< a (removable) constraint learnt by some other means          */
 		max_value         = learnt_other
 	};
+	struct Set {
+		Set() : m(0) {}
+		bool inSet(Type t) const { return (m & (uint32(1)<<t)) != 0; }
+		void addSet(Type t)      { m |= (uint32(1)<<t); }
+		uint32 m;
+	};
 };  
 typedef Constraint_t::Type ConstraintType;
+typedef Constraint_t::Set  TypeSet;
+//! Type storing a constraint's activity.
+struct Activity {
+	enum { LBD_SHIFT = 7, MAX_LBD = (1 << LBD_SHIFT)-1, MAX_ACT = (1 << (32-LBD_SHIFT))-1 };
+	Activity(uint32 act, uint32 lbd) : rep( (act << LBD_SHIFT) | lbd ) {}
+	uint32 activity() const { return rep >> LBD_SHIFT; }
+	uint32 lbd()      const { return rep & uint32(MAX_LBD); }
+	void   bumpAct()        { if (rep < uint32(MAX_ACT<<LBD_SHIFT)) rep += (1<<LBD_SHIFT); }
+	void   setLbd(uint32 x) { rep &= ~uint32(MAX_LBD); rep |= x; }
+	uint32 rep;
+};
 
 //! Base class for (boolean) constraints to be used in a Solver.
 /*!
@@ -67,17 +84,13 @@ typedef Constraint_t::Type ConstraintType;
  */
 class Constraint {
 public:
-	/*!
-	 * Type used as return type for Constraint::propagate.
-	 */
+	//! Type used as return type for Constraint::propagate.
 	struct PropResult {
 		explicit PropResult(bool a_ok = true, bool a_keepWatch = true) : ok(a_ok), keepWatch(a_keepWatch) {}
 		bool ok;         /**< true if propagation completes without conflict     */
 		bool keepWatch;  /**< true if constraint wants to keep the current watch */
 	};
-	
 	Constraint();
-
 	//! Returns a clone of this and adds necessary watches to the given solver.
 	/*!
 	 * The function shall create and return a copy of this constraint
@@ -192,17 +205,6 @@ public:
 	 */
 	virtual bool locked(const Solver& s) const = 0;
 
-	//! Type storing a constraint's activity.
-	struct Activity {
-		uint32 score()    const { return score_; }
-		uint32 lbd()      const { return lbd_; }
-		uint32 scoreAct() const { return score(); }
-		uint32 scoreLbd() const { return uint32(128)-lbd(); }
-		uint32 scoreBoth()const { return (score()+1) * scoreLbd(); }
-		uint32 score_ : 25;
-		uint32 lbd_   : 7;
-	};
-	
 	//! Returns the activity of the constraint.
 	/*!
 	 * \note A solver uses the activity-value in order to establish an ordering
@@ -214,18 +216,16 @@ public:
 	
 	//! Asks the constraint to decrease its activity.
 	virtual void decreaseActivity();
-
-	//! Shall return true if this constraint is satisfied w.r.t the current assignment.
+	
+	//! Shall return 0 if either !t.inSet(type()) or this constraint is satisfied w.r.t the current assignment.
 	/*!
-	 * If this function returns false, freeLits shall contain some (or all) currently
-	 * unassigned literals of this constraint.
+	 * If this constraint is currently not satisfied and t.inSet(type()), shall return type()
+	 * and freeLits shall contain all currently unassigned literals of this constraint.
 	 */
-	virtual bool isSatisfied(const Solver& s, LitVec& freeLits) = 0;
+	virtual uint32 isOpen(const Solver& s, const TypeSet& t, LitVec& freeLits) = 0;
 protected:
 	~LearntConstraint();
 };
-inline LearntConstraint::Activity makeActivity(uint32 a_act, uint32 a_lbd) { LearntConstraint::Activity a = {a_act, a_lbd}; return a; }
-inline LearntConstraint::Activity maxActivity() { return makeActivity((1u<<25)-1, 1); }
 
 //! Base class for all post propagators.
 /*!
@@ -376,13 +376,13 @@ private:
  * 
  * From the 64-bits the first 2-bits encode the type stored:
  *  - 00: Pointer to constraint
- *  - 01: binary constraint (i.e. one literal stored in the highest 31 bits)
- *  - 11: ternary constraint (i.e. two literals stored in the remaining 62 bits). 
+ *  - 01: ternary constraint (i.e. two literals stored in the remaining 62 bits). 
+ *  - 10: binary constraint (i.e. one literal stored in the highest 31 bits)
  * 
  */
 class Antecedent {
 public:
-	enum Type { generic_constraint = 0, binary_constraint = 1, ternary_constraint = 3};
+	enum Type { generic_constraint = 0, ternary_constraint = 1, binary_constraint = 2};
 	//! Creates a null Antecedent. 
 	/*!
 	 * \post: isNull() == true && type == generic_constraint
@@ -428,6 +428,9 @@ public:
 		return Type( data_ & 3 );
 	}
 	
+	//! Returns true if the antecedent is a learnt nogood.
+	bool learnt() const { return data_ && (data_ & 3u) == 0 && constraint()->type() != Constraint_t::static_constraint; }
+	
 	//! Extracts the constraint-pointer stored in this object.
 	/*!
 	 * \pre type() == generic_constraint
