@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2006-2010, Benjamin Kaufmann
+// Copyright (c) 2006-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -33,58 +33,67 @@
  */
 namespace Clasp { 
 
+template <class T>
+struct Range {
+	Range(T x, T y) : lo(x), hi(y) { if (x > y)  { hi = x;  lo = y; } }
+	T clamp(T val) const { 
+		if (val < lo) return lo;
+		if (val > hi) return hi;
+		return val;
+	}
+	T lo;
+	T hi;
+};
+typedef Range<uint32> Range32;
+	
 //! Aggregates restart-parameters to configure restarts during search.
 /*!
  * \see ScheduleStrategy
  */
 struct RestartParams {
-	enum Type { 
-		default_restarts = 0, /**< Restart based on schedule. */
-		local_restarts   = 1, /**< Restart if number of conflicts in *one* branch exceed threshold. */
-		dynamic_restarts = 2  /**< Dynamic restarts. Ignore schedule. */
-	};
-	RestartParams() : sched(), cirBump(9973), cir(0), type(default_restarts), bounded(false), resetOnModel(false) {}
-	void initDynamic(uint32 base, double xLbd, double xCfl) {
-		sched.base = base;
-		sched.grow = xLbd;
-		sched.outer= static_cast<uint64>(xCfl * 1e8);
-		type       = dynamic_restarts;
-	}
-	void disable() { sched = ScheduleStrategy::none();  }
+	RestartParams() : sched(), counterRestart(0), counterBump(9973), shuffle(0), shuffleNext(0), cntLocal(0), dynRestart(0), bndRestart(0), rstRestart(0) {}
+	void disable();
+	bool dynamic()     const { return dynRestart != 0; }
+	bool local()       const { return cntLocal   != 0; }
+	bool bounded()     const { return bndRestart != 0; }
+	bool resetOnModel()const { return rstRestart != 0; }
 	ScheduleStrategy sched; /**< Restart schedule to use. */
-	uint32 cirBump;         /**< Bump factor for counter implication restarts. */
-	uint8  cir;             /**< Apply counter implication bump every cir restarts (0: disable). */
-	uint8  type;            /**< Type of restarts. */
-	bool   bounded;         /**< Allow (bounded) restarts after first solution was found. */
-	bool   resetOnModel;    /**< Repeat restart strategy after each solution. */
+	uint16 counterRestart;  /**< Apply counter implication bump every counterRestart restarts (0: disable). */
+	uint16 counterBump;     /**< Bump factor for counter implication restarts. */
+	uint32 shuffle    :14;  /**< Shuffle program after shuffle restarts (0: disable). */
+	uint32 shuffleNext:14;  /**< Re-Shuffle program every shuffleNext restarts (0: disable). */
+	uint32 cntLocal   : 1;  /**< Count conflicts globally or relative to current branch? */
+	uint32 dynRestart : 1;  /**< Dynamic restarts enabled? */
+	uint32 bndRestart : 1;  /**< Allow (bounded) restarts after first solution was found. */
+	uint32 rstRestart : 1;  /**< Repeat restart seq. after each solution. */
 };
 
-
 //! Aggregates parameters for the nogood deletion heuristic used during search.
 struct ReduceParams {
-	ReduceParams() : cflSched(ScheduleStrategy::none()), growSched(ScheduleStrategy::null()), baseFrac(3.0f), remFrac(0.75f)
-		, dbGrow(1.1f), dbMaxGrow(3.0f)
-		, base(5000), baseMin(10), baseMax(UINT32_MAX)
-		, reduceOnRestart(false), estimate(false) {}
-	
-	void disable() {
-		cflSched = ScheduleStrategy::none();
-		growSched= ScheduleStrategy::none();
+	ReduceParams() : cflSched(ScheduleStrategy::none()), growSched(ScheduleStrategy::def())
+		, fInit(1.0f/3.0f)
+		, fMax(3.0f)
+		, fGrow(1.1f)
+		, initRange(10, UINT32_MAX)
+		, maxRange(UINT32_MAX) {}
+	void   disable();
+	uint32 getBase(const SharedContext& ctx)   const;
+	uint32 initLimit(const SharedContext& ctx) const { return getLimit(getBase(ctx), fInit, initRange); }
+	uint32 maxLimit(const SharedContext& ctx)  const { 
+		return getLimit(getBase(ctx), fMax, Range32(initLimit(ctx), maxRange)); 
 	}
-	uint32 initDbSize() const;
-	uint32 maxDbSize()  const;
-	ScheduleStrategy cflSched; /**< Conflict-based deletion schedule.               */
-	ScheduleStrategy growSched;/**< Growth-based deletion schedule.                 */
-	ReduceStrategy   strategy; /**< Strategy to apply during nogood deletion.       */
-	float            baseFrac; /**< Initial db size is base / baseFrac.             */
-	float            remFrac;  /**< Fraction of nogoods to remove ([0.0,1.0]).      */
-	float            dbGrow;   /**< Growth factor for db.                           */
-	float            dbMaxGrow;/**< Stop growth once db size is > base*dbMaxGrow.   */
-	uint32           base;     /**< Initial problem size estimate.                  */
-	uint32           baseMin;  /**< Minimal initial db size.                        */
-	uint32           baseMax;  /**< Maximal initial db size.                        */
-	bool             reduceOnRestart; /**< Delete some nogoods on each restart.     */
-	bool             estimate; /**< Use estimate of problem complexity in db init.  */
+	bool   estimate() const { return strategy.estimate != 0; }
+	float  fReduce()  const { return strategy.fReduce / 100.0f; }
+	float  fRestart() const { return strategy.fRestart/ 100.0f; }
+	static uint32 getLimit(uint32 base, double f, const Range<uint32>& r);
+	ScheduleStrategy cflSched;   /**< Conflict-based deletion schedule.               */
+	ScheduleStrategy growSched;  /**< Growth-based deletion schedule.                 */
+	ReduceStrategy   strategy;   /**< Strategy to apply during nogood deletion.       */
+	float            fInit;      /**< Initial limit. X = P*fInit clamped to initRange.*/
+	float            fMax;       /**< Maximal limit. X = P*fMax  clamped to maxRange. */
+	float            fGrow;      /**< Growth factor for db.                           */
+	Range32          initRange;  /**< Allowed range for initial limit.                */
+	uint32           maxRange;   /**< Allowed range for maximal limit: [initRange.lo,maxRange]*/
 };
 
 //! Type for holding global solve limits.
@@ -100,11 +109,11 @@ struct SolveLimits {
 
 //! Type for holding pre-solve options.
 struct InitParams {
-	InitParams() : randRuns(0), randConf(0), initLook(0), lookType(0) {}
+	InitParams() : randRuns(0), randConf(0), lookType(0), lookOps(0) {}
 	uint16 randRuns; /*!< Number of initial randomized-runs. */
 	uint16 randConf; /*!< Number of conflicts comprising one randomized-run. */
-	uint16 initLook; /*!< Number of initial lookahead operations. */
-	uint16 lookType; /*!< Type of initial lookahead operations. */
+	uint16 lookType; /*!< Type of lookahead operations. */
+	uint16 lookOps;  /*!< Max. number of lookahead operations (0: no limit). */
 };
 ///////////////////////////////////////////////////////////////////////////////
 // Parameter-object for the solve function
@@ -119,7 +128,6 @@ struct SolveParams {
 	/*!
 	 * The following parameters are used:
 	 * restart      : quadratic: 100*1.5^k / no restarts after first solution
-	 * shuffle      : disabled
 	 * deletion     : initial size: vars()/3, grow factor: 1.1, max factor: 3.0, do not reduce on restart
 	 * randomization: disabled
 	 * randomProp   : 0.0 (disabled)
@@ -129,18 +137,6 @@ struct SolveParams {
 	RestartParams restart;
 	ReduceParams  reduce;
 	InitParams    init;
-	mutable SolveLimits  limits;
-	
-	//! Sets the shuffle-parameters to use during search.
-	/*!
-	 * \param first   Shuffle program after first restarts.
-	 * \param next    Re-Shuffle program every next restarts.
-	 * \note          If first is equal to 0, shuffling is disabled.
-	 */
-	void setShuffleParams(uint32 first, uint32 next) {
-		shuffleFirst_ = first;
-		shuffleNext_  = next;
-	}
 
 	//! Sets the randomization-parameters to use during path initialization.
 	/*!
@@ -165,15 +161,16 @@ struct SolveParams {
 	uint32  randRuns()          const { return init.randRuns; }
 	uint32  randConflicts()     const { return init.randConf; }
 	double  randomProbability() const { return randFreq_;     }
-	uint32  shuffleBase()       const { return shuffleFirst_; }
-	uint32  shuffleNext()       const { return shuffleNext_;  }
 private:
 	double randFreq_;
-	uint32 shuffleFirst_;
-	uint32 shuffleNext_;
 };
-
-
+//! Object holding options for one solver instance.
+struct SolverConfig {
+	explicit    SolverConfig(Solver& s) : solver(&s) {}
+	void        initFrom(const SolverConfig& other);
+	Solver*     solver;
+	SolveParams params;
+};
 ///////////////////////////////////////////////////////////////////////////////
 // Basic solve functions
 ///////////////////////////////////////////////////////////////////////////////
@@ -184,13 +181,14 @@ private:
  * \relates Solver
  * \param ctx The context containing the problem.
  * \param p   The solve parameters to use.
+ * \param lim Optional solve limit. 
  *
  * \return
  *  - true:  if the search stopped before the search-space was exceeded.
  *  - false: if the search-space was completely examined.
  * 
  */
-bool solve(SharedContext& ctx, const SolveParams& p);
+bool solve(SharedContext& ctx, const SolveParams& p, const SolveLimits& lim = SolveLimits());
 
 //! Basic sequential search under assumptions.
 /*!
@@ -202,18 +200,35 @@ bool solve(SharedContext& ctx, const SolveParams& p);
  * \param ctx The context containing the problem.
  * \param p   The solve parameters to use.
  * \param assumptions The list of initial unit-assumptions.
+ * \param lim Optional solve limit.
  *
  * \return
  *  - true:  if the search stopped before the search-space was exceeded.
  *  - false: if the search-space was completely examined.
  * 
  */
-bool solve(SharedContext& ctx, const SolveParams& p, const LitVec& assumptions);
+bool solve(SharedContext& ctx, const SolveParams& p, const LitVec& assumptions, const SolveLimits& lim = SolveLimits());
 
 ///////////////////////////////////////////////////////////////////////////////
 // General solve
 ///////////////////////////////////////////////////////////////////////////////
-
+struct SolveEvent_t { enum Type {progress_msg = 1, progress_state = 2, progress_path = 3}; };
+struct SolveMsgEvent : SolveEvent {
+	SolveMsgEvent(const Solver& s, const char* m) : SolveEvent(s, SolveEvent_t::progress_msg), msg(m) {}
+	const char* msg;
+};
+struct SolveStateEvent : SolveEvent {
+	SolveStateEvent(const Solver& s, const char* m, double exitTime = -1.0) : SolveEvent(s, SolveEvent_t::progress_state), state(m), time(exitTime) {}
+	const char* state; // state that is entered or left
+	double      time;  // < 0: enter, else exit
+};
+struct SolvePathEvent : SolveEvent {
+	enum EventType { event_none = 0, event_deletion = 'D', event_grow = 'G', event_model = 'M', event_restart = 'R' };
+	SolvePathEvent(const Solver& s, EventType t, uint64 cLim, uint32 lLim) : SolveEvent(s, SolveEvent_t::progress_path), cLimit(cLim), lLimit(lLim), evType(t) {}
+	uint64    cLimit; // next conflict limit
+	uint32    lLimit; // next learnt limit
+	EventType evType; // type of path event
+};
 //! Interface for solve algorithms.
 /*!
  * \ingroup solver
@@ -223,19 +238,22 @@ bool solve(SharedContext& ctx, const SolveParams& p, const LitVec& assumptions);
  */
 class SolveAlgorithm {
 public:
-	explicit SolveAlgorithm();
+	/*!
+	 * \param lim    An optional solve limit applied in solve().
+	 */
+	explicit SolveAlgorithm(const SolveLimits& limit = SolveLimits());
 	virtual ~SolveAlgorithm();
-	
+
 	//! Force termination of current solve process.
 	/*!
 	 * Shall return true if termination is supported, otherwise false.
 	 */
 	virtual bool   terminate() = 0;
-
+	virtual bool   hasErrors() { return false; }
 	//! Runs the solve algorithm.
 	/*!
 	 * \param ctx    A fully initialized context object containing the problem.
-	 * \param p      The solve parameters for the master solver.
+	 * \param p      The solve parameters for the master solver ctx.master()
 	 * \param assume A list of initial unit-assumptions.
 	 *
 	 * \return
@@ -246,28 +264,29 @@ public:
 	 * The use of assumptions allows for incremental solving. Literals contained
 	 * in assumptions are assumed to be true during search but are undone before solve returns.
 	 */
-	bool solve(SharedContext& ctx, const SolveParams& p, LitVec assume);
+	bool solve(SharedContext& ctx, const SolveParams& p, const LitVec& assume);
 protected:
 	//! The default implementation simply forwards the call to the enumerator.
-	virtual bool backtrackFromModel(Solver& s);
-	//! The default implementation simply forwards the call to the enumerator.
-	virtual void reportProgress(int type, Solver& s, uint64 maxCfl, uint32 maxLearnt);
-	virtual bool doSolve(Solver& s, const SolveParams& p, const LitVec& assume) = 0;
-	bool     initPath(Solver& s,  const LitVec& gp, InitParams& params);
-	ValueRep solvePath(Solver& s, const SolveParams& p);
+	virtual bool  backtrackFromModel(Solver& s);
+	virtual bool  doSolve(Solver& s, const SolveParams& p) = 0;
+	bool          initPath(Solver& s,  const LitVec& gp, InitParams& params);
+	ValueRep      solvePath(Solver& s, const SolveParams& p, SolveLimits& inOut);
+	const LitVec& getInitialPath()     const { return assumptions_; }
+	const SolveLimits& getSolveLimits()const { return limits_; }
+	void               setSolveLimits(const SolveLimits& x) { limits_ = x; }
 private:
 	SolveAlgorithm(const SolveAlgorithm&);
 	SolveAlgorithm& operator=(const SolveAlgorithm&);
+	SolveLimits    limits_;
+	LitVec         assumptions_;
 };
-
 //! A basic algorithm for single-threaded sequential solving.
 class SimpleSolve : public SolveAlgorithm {
 public:
-	SimpleSolve() : SolveAlgorithm() {}
+	SimpleSolve(const SolveLimits& lim) : SolveAlgorithm(lim) {}
 	bool   terminate();
 private:
-	bool   doSolve(Solver& s, const SolveParams& p, const LitVec& assume);
+	bool   doSolve(Solver& s, const SolveParams& p);
 };
-
 }
 #endif
