@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2006-2011, Benjamin Kaufmann
+// Copyright (c) 2006-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -24,8 +24,6 @@
 #pragma once
 #endif
 
-#include <clasp/util/atomic.h>
-
 #define STRING2(x) #x
 #define STRING(x) STRING2(x)
 
@@ -70,6 +68,9 @@ typedef uintptr_t   uintp;
 #ifndef UINT64_MAX
 #define UINT64_MAX (~uint64(0))
 #endif
+#ifndef INT64_MAX
+#define INT64_MAX ((int64)(UINT64_MAX >> 1))
+#endif
 #ifndef UINTP_MAX
 #define UINTP_MAX (~uintp(0))
 #endif
@@ -84,26 +85,6 @@ typedef uintptr_t   uintp;
 #define store_clear_bit(x,n) ( (x) &= ~BIT_MASK((x),(n)) )
 #define store_toggle_bit(x,n)( (x) ^=  BIT_MASK((x),(n)) )
 
-// effect: T temp; a |= mask; return temp
-template <class T>
-inline T fetch_and_or(std::atomic<T>& a, T mask) {
-	T x;
-	do {
-		x = a;
-	} while (a.compare_and_swap(x|mask, x) != x);
-	return x;
-}
-
-// effect: T temp; a &= mask; return temp
-template <class T>
-inline T fetch_and_and(std::atomic<T>& a, T mask) {
-	T x;
-	do {
-		x = a;
-	} while (a.compare_and_swap(x&mask, x) != x);
-	return x;
-}
-
 // return true if bit n in x is set
 #define test_bit(x,n)  ( ((x) & BIT_MASK((x),(n))) != 0 )
 
@@ -123,6 +104,6 @@ bool aligned(void* mem) {
 template <bool> struct static_assertion;
 template <>     struct static_assertion<true> {};
 
-#define static_assert(x) (void)sizeof(static_assertion< (x) >)
+#define static_assert(x, message) (void)sizeof(static_assertion< (x) >)
 
 #endif
