@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2006-2011, Benjamin Kaufmann
+// Copyright (c) 2006-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -53,6 +53,13 @@ static uint32 levelAbstraction(const Solver& s, Literal p) {
 	return abstr_p;
 }
 
+void* alloc(uint32 size) {
+	#pragma message TODO("replace with CACHE_LINE_ALIGNED alloc")
+	return ::operator new(size);
+}
+void free(void* mem) {
+	::operator delete(mem);
+}
 
 } // namespace Detail
 
@@ -60,14 +67,13 @@ static uint32 levelAbstraction(const Solver& s, Literal p) {
 // SharedLiterals
 /////////////////////////////////////////////////////////////////////////////////////////
 SharedLiterals* SharedLiterals::newShareable(const Literal* lits, uint32 size, ConstraintType t, uint32 numRefs) {
-	#pragma message TODO("replace with CACHE_LINE_ALIGNED alloc")
-	void* m = ::operator new(sizeof(SharedLiterals)+(size*sizeof(Literal)));
+	void* m = Detail::alloc(sizeof(SharedLiterals)+(size*sizeof(Literal)));
 	return new (m) SharedLiterals(lits, size, t, numRefs);
 }
 
 SharedLiterals::SharedLiterals(const Literal* a_lits, uint32 size, ConstraintType t, uint32 refs) 
 	: size_type_( (size << 2) + t ) {
-	refCount_ = refs;
+	refCount_ = std::max(uint32(1),refs);
 	std::memcpy(lits_, a_lits, size*sizeof(Literal));
 }
 
@@ -96,7 +102,7 @@ uint32 SharedLiterals::simplify(Solver& s) {
 void SharedLiterals::release() {
 	if (--refCount_ == 0) {
 		this->~SharedLiterals();
-		::operator delete(this);
+		Detail::free(this);
 	}
 }
 SharedLiterals* SharedLiterals::share() {
@@ -179,7 +185,7 @@ ClauseCreator::ClauseCreator(Solver* s)
 void ClauseCreator::init(ConstraintType t) {
 	assert(!solver_ || solver_->decisionLevel() == 0 || t != Constraint_t::static_constraint);
 	literals_.clear();
-	extra_     = ClauseInfo().setType(t);
+	extra_     = ClauseInfo(t);
 	fwLev_     = 0;
 	swLev_     = 0;
 }
@@ -290,13 +296,11 @@ ClauseCreator::Status ClauseCreator::status(const Solver& s, const Watches& w, u
 
 ClauseCreator::Result ClauseCreator::end() {
 	assert(solver_);
-	Result ret(0, true);
+	Result ret(0);
 	Solver& s     = *solver_;
 	uint32  flags = clause_known_order | clause_not_sat | clause_not_conflict;
-	if (extra_.type() != Constraint_t::static_constraint && !isSentinel(s.sharedContext()->tagLiteral())) {
-		if (std::find(literals_.begin(), literals_.end(), ~s.sharedContext()->tagLiteral()) != literals_.end()) {
-			extra_.tag();
-		}
+	if (extra_.learnt() && !isSentinel(s.sharedContext()->tagLiteral())) {
+		extra_.setTagged(std::find(literals_.begin(), literals_.end(), ~s.sharedContext()->tagLiteral()) != literals_.end());
 	}
 	return ClauseCreator::create(s, literals_, flags, extra_);
 }
@@ -304,14 +308,15 @@ ClauseCreator::Result ClauseCreator::end() {
 ClauseCreator::Status ClauseCreator::initWatches(Solver& s, LitVec& lits, bool allFree) {
 	uint32 size = static_cast<uint32>(lits.size());
 	if (size < 2) return size == 1 ? status_unit : status_empty;
-	uint32 fw, sw;
 	if (allFree) {
-		if (s.strategies().randomWatches) {
+		uint32 fw = 0, sw = 1;
+		uint32 w  = s.strategies().initWatches;
+		if (w == SolverStrategies::watch_rand) {
 			RNG& r = s.strategies().rng;
 			fw = r.irand(size);
 			while ( (sw = r.irand(size)) == fw) {/*intentionally empty*/}
 		}
-		else {
+		else if (w == SolverStrategies::watch_least) {
 			uint32 watch[2] = {0, size-1};
 			uint32 count[2] = {s.numWatches(~lits[0]), s.numWatches(~lits[size-1])};
 			uint32 maxCount = count[0] < count[1];
@@ -346,7 +351,7 @@ ClauseCreator::Status ClauseCreator::initWatches(Solver& s, LitVec& lits, bool a
 ClauseHead* ClauseCreator::newProblemClause(Solver& s, LitVec& lits, const ClauseInfo& e, uint32 flags) {
 	ClauseHead* ret;
 	initWatches(s, lits, true);
-	if (lits.size() <= Clause::MAX_SHORT_LEN || !s.sharedContext()->shareConstraints()) {
+	if (lits.size() <= Clause::MAX_SHORT_LEN || !s.sharedContext()->physicalShareProblem()) {
 		ret = Clause::newClause(s, &lits[0], (uint32)lits.size(), e);
 	}
 	else {
@@ -376,66 +381,77 @@ ClauseHead* ClauseCreator::newLearntClause(Solver& s, LitVec& lits, const Clause
 		sharedPtr.clause = 0;
 	}
 	if ( (flags & clause_no_add) == 0 ) {
-		s.addLearnt(ret, (uint32)lits.size(), e.tagged());
+		s.addLearnt(ret, (uint32)lits.size(), e.type());
 	}
 	return ret;
 }
 
+ClauseHead* ClauseCreator::newUnshared(Solver& s, SharedLiterals* clause, const Watches& w, const ClauseInfo& e) {
+	LitVec temp; temp.reserve(clause->size());
+	temp.assign(w.lits, w.lits+2);
+	for (const Literal* x = clause->begin(), *end = clause->end(); x != end; ++x) {
+		if (Detail::levelAbstraction(s, *x) > 0 && *x != temp[0] && *x != temp[1]) {
+			temp.push_back(*x);
+		}
+	}
+	return Clause::newClause(s, &temp[0], (uint32)temp.size(), e);
+}
+
 ClauseCreator::Result ClauseCreator::create(Solver& s, LitVec& lits, uint32 flags, const ClauseInfo& extra) {
-	assert(s.decisionLevel() == 0 || extra.type() != Constraint_t::static_constraint);
+	assert(s.decisionLevel() == 0 || extra.learnt());
 	Result ret;
 	if (lits.size() > 1) {
 		Watches w(s, lits, (flags & clause_known_order) != 0);
-		Status x = status(s, w, flags);
+		Status x   = status(s, w, flags);
+		ret.status = x;
 		if (x == status_subsumed || x == status_conflicting) { 
-			ret.ok = (x == status_subsumed);
 			return ret;
 		}
-		if (extra.type() == Constraint_t::static_constraint && s.satPrepro()) {
-			ret.ok = s.satPrepro()->addClause(lits);
+		if (!extra.learnt() && s.satPrepro()) {
+			if (!s.satPrepro()->addClause(lits)) { ret.status = status_conflicting; }
 			return ret;
 		}		
-		s.strategies().heuristic->newConstraint(s, &lits[0], lits.size(), extra.type());
-		bool   unit = (x & status_unit) != 0;
-		uint32 impL = (unit ? s.level(lits[1].var()) : UINT32_MAX);
+		s.heuristic()->newConstraint(s, &lits[0], lits.size(), extra.type());
 		if (lits.size() > 3 || extra.tagged() || (flags&clause_explicit) != 0) {
-			ret.local = (extra.type() == Constraint_t::static_constraint)
-					? newProblemClause(s, lits, extra, flags)
-					: newLearntClause(s, lits, extra, flags);
-			ret.ok = !unit || s.force(lits[0], impL, Antecedent(ret.local));
-		}
-		else if (lits.size() == 3) {
-			s.addTernary(lits[0], lits[1], lits[2], extra);
-			ret.ok = !unit || s.force(lits[0], impL, Antecedent(~lits[1], ~lits[2]));
+			ret.local = extra.learnt() ? newLearntClause(s, lits, extra, flags) : newProblemClause(s, lits, extra, flags);
 		}
 		else {
-			s.addBinary(lits[0], lits[1], extra);
-			ret.ok = !unit || s.force(lits[0], impL, Antecedent(~lits[1]));
+			s.addShort(&lits[0], lits.size(), extra);
+		}
+		if ((x & status_unit) != 0) {
+			Antecedent ante(ret.local);
+			if (!ret.local){ ante = lits.size() == 3 ? Antecedent(~lits[1], ~lits[2]) : Antecedent(~lits[1]); }
+			if (!s.force(lits[0], s.level(lits[1].var()), ante)) { ret.status = status_conflicting; }
 		}
 	}
 	else {
 		Literal imp = !lits.empty()	? lits[0] : negLit(0);
-		ret.ok = s.addUnary(imp, extra.type());
+		if (!s.addUnary(imp, extra.type())) { ret.status = status_conflicting; }
 	}	
 	return ret;
 }
 
 
-ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags, const ClauseInfo& e) {
+ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags, ConstraintType t) {
 	assert(!s.hasConflict() && "ClauseCreator::integrate() - precondition violated!");
 	Detail::Sink shared( (modeFlags & clause_no_release) == 0 ? clause : 0);
 	// determine state of clause
 	Result  result;
 	Watches w(s, clause->begin(), clause->end(), (modeFlags & clause_known_order) != 0);
 	Status  stat( status(s, w, modeFlags) );
+	result.status = stat;
 	if (stat != status_subsumed && (stat != status_conflicting || (modeFlags & clause_not_conflict) == 0)) {
-		uint32 impSize  = (modeFlags & clause_explicit) != 0 || !s.sharedContext()->updateShortImplications() ? 1 : 3;
-		if (w.simpSize > Clause::MAX_SHORT_LEN) {
+		ClauseInfo e(t);
+		s.heuristic()->newConstraint(s, clause->begin(), clause->size(), t);
+		uint32 impSize  = (modeFlags & clause_explicit) != 0 || !s.sharedContext()->allowImplicit(t) ? 1 : 3;
+		if (w.simpSize > Clause::MAX_SHORT_LEN && s.sharedContext()->physicalShare(t)) {
 			result.local  = Clause::newShared(s, clause, e, w.lits, shared.clause == 0);
 			shared.clause = 0;
 		}
 		else if (w.simpSize > impSize) {
-			result.local  = Clause::newClause(s, w.lits, w.simpSize, e);
+			result.local  = w.simpSize <= Clause::MAX_SHORT_LEN
+			              ? Clause::newClause(s, w.lits, w.simpSize, e)
+			              : newUnshared(s, clause, w, e);
 		}
 		else {
 			// unary clause or implicitly shared via binary/ternary implication graph;
@@ -444,40 +460,38 @@ ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause
 			s.stats.addLearnt(w.simpSize, e.type());
 			modeFlags    |= clause_no_add;
 		}
-		if ((modeFlags & clause_no_add) == 0) { s.addLearnt(result.local, w.simpSize); }
+		if ((modeFlags & clause_no_add) == 0) { s.addLearnt(result.local, w.simpSize, e.type()); }
 	}
 	if ((stat & status_unit) != 0) {
 		Antecedent ante = result.local ? Antecedent(result.local) : Antecedent(~w.lits[1], ~w.lits[2]);
-		uint32 DL       = s.decisionLevel();
 		uint32 impLevel = s.level(w.lits[1].var());
-		result.ok       = s.force(w.lits[0], impLevel, ante);
-		if (s.stats.parallel){
-			s.stats.parallel->imps  += 1;
-			s.stats.parallel->jumps += (DL - s.decisionLevel());
+		result.status   = s.force(w.lits[0], impLevel, ante) ? stat : status_conflicting;
+		if (result.local && (modeFlags & clause_int_lbd) != 0 && result.status != status_conflicting) {
+			result.local->lbd(s.updateLearnt(negLit(0), clause->begin(), clause->end(), result.local->lbd(), true));
 		}
 	}
-	else { result.ok  = stat != status_conflicting; }
 	return result;
 }
 ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags) { 
-	return integrate(s, clause, modeFlags, ClauseInfo().setType(clause->type())); 
+	return integrate(s, clause, modeFlags, clause->type());
 }
 /////////////////////////////////////////////////////////////////////////////////////////
 // Clause
 /////////////////////////////////////////////////////////////////////////////////////////
-
-void* Clause::alloc(Solver& s, uint32 lits) {
+void* Clause::alloc(Solver& s, uint32 lits, bool learnt) {
 	if (lits <= Clause::MAX_SHORT_LEN) {
+		if (learnt) { s.addLearntBytes(32); }
 		return s.allocSmall();
 	}
-	#pragma message TODO("replace with CACHE_LINE_ALIGNED alloc") 
 	uint32 extra = std::max((uint32)ClauseHead::HEAD_LITS, lits) - ClauseHead::HEAD_LITS; 
-	return ::operator new(sizeof(Clause) + (extra)*sizeof(Literal));
+	uint32 bytes = sizeof(Clause) + (extra)*sizeof(Literal);
+	if (learnt) { s.addLearntBytes(bytes); }
+	return Detail::alloc(bytes);
 }
 
-ClauseHead* Clause::newClause(Solver& s, const Literal* lits, uint32 size, const ClauseInfo& e) {
-	assert(size >= 2);
-	return new (alloc(s, size)) Clause(s, e, lits, size, size, false);
+ClauseHead* Clause::newClause(void* mem, Solver& s, const Literal* lits, uint32 size, const ClauseInfo& extra) {
+	assert(size >= 2 && mem);
+	return new (mem) Clause(s, extra, lits, size, size, false);
 }
 
 ClauseHead* Clause::newShared(Solver& s, SharedLiterals* shared_lits, const ClauseInfo& e, const Literal* lits, bool addRef) {
@@ -486,13 +500,7 @@ ClauseHead* Clause::newShared(Solver& s, SharedLiterals* shared_lits, const Clau
 
 ClauseHead* Clause::newContractedClause(Solver& s, const LitVec& lits, const ClauseInfo& e, LitVec::size_type tailStart, bool extend) {
 	assert(lits.size() >= 2);
-	return new (alloc(s, (uint32)lits.size())) Clause(s, e, &lits[0], static_cast<uint32>(lits.size()), static_cast<uint32>(tailStart), extend);
-}
-
-ClauseHead* Clause::newShortClause(Solver& s, void* mem, const Literal* lits, uint32 size, const ClauseInfo& extra) {
-	assert(size >= 2);
-	assert(size <= static_cast<uint32>(Clause::MAX_SHORT_LEN));
-	return new (mem ? mem : s.allocSmall()) Clause(s, extra, lits, size, size, false);
+	return new (alloc(s, (uint32)lits.size(), e.learnt())) Clause(s, e, &lits[0], static_cast<uint32>(lits.size()), static_cast<uint32>(tailStart), extend);
 }
 
 Clause::Clause(Solver& s, const ClauseInfo& e, const Literal* a_lits, uint32 size, uint32 tail, bool extend) 
@@ -538,19 +546,20 @@ Clause::Clause(Solver& s, const Clause& other) : ClauseHead(ClauseInfo()) {
 }
 
 Constraint* Clause::cloneAttach(Solver& other) {
-	assert(type() == Constraint_t::static_constraint);
-	return new (alloc(other, Clause::size())) Clause(other, *this);
+	assert(!learnt());
+	return new (alloc(other, Clause::size(), false)) Clause(other, *this);
 }
 
 void Clause::destroy(Solver* s, bool detachFirst) {
-	if (detachFirst && s) { 
-		Clause::detach(*s);
+	if (s) { 
+		if (detachFirst) { Clause::detach(*s); }
+		if (learnt())    { s->freeLearntBytes(computeAllocSize()); }
 	}
 	void* mem   = static_cast<Constraint*>(this);
 	bool  small = isSmall();
 	this->~Clause();
-	if (!small) { ::operator delete(mem); }
-	else if (s) { s->freeSmall(mem);      }
+	if (!small) { Detail::free(mem); }
+	else if (s) { s->freeSmall(mem); }
 }
 
 void Clause::detach(Solver& s) {
@@ -563,6 +572,19 @@ void Clause::detach(Solver& s) {
 	ClauseHead::detach(s);
 }
 
+uint32 Clause::computeAllocSize() const {
+	if (isSmall()) { return 32; }
+	uint32 rt = sizeof(Clause) - (ClauseHead::HEAD_LITS*sizeof(Literal));
+	uint32 sz = data_.local.size();
+	uint32 nw = contracted() + strengthened();
+	if (nw != 0u) {
+		const Literal* eoc = head_ + sz;
+		do { nw -= eoc++->watched(); } while (nw); 
+		sz = eoc - head_;
+	}
+	return rt + (sz*sizeof(Literal));
+}
+
 bool Clause::updateWatch(Solver& s, uint32 pos) {
 	uint32 idx = data_.local.idx;
 	if (!isSmall()) {
@@ -587,7 +609,6 @@ bool Clause::updateWatch(Solver& s, uint32 pos) {
 }
 
 void Clause::reason(Solver& s, Literal p, LitVec& out) {
-	ClauseHead::bumpActivity();
 	LitVec::size_type i = out.size();
 	out.push_back(~head_[p == head_[0]]);
 	if (!isSentinel(head_[2])) {
@@ -601,9 +622,9 @@ void Clause::reason(Solver& s, Literal p, LitVec& out) {
 			do { out.push_back(~*r); } while (!r++->watched());
 		}
 	}
-	if (s.strategies().updateLbd && ClauseHead::type() != Constraint_t::static_constraint) {
-		uint32 n = s.computeLbd(p, &out[0]+i, &out[0]+out.size());
-		if (n < info_.data.lbd) { setLbd( n ); }
+	if (learnt()) { 
+		ClauseHead::bumpActivity();
+		setLbd(s.updateLearnt(p, &out[0]+i, &out[0]+out.size(), lbd(), !hasLbd())); 
 	}
 }
 
@@ -640,7 +661,7 @@ bool Clause::isReverseReason(const Solver& s, Literal p, uint32 maxL, uint32 max
 }
 
 bool Clause::simplify(Solver& s, bool reinit) {
-	assert(s.decisionLevel() == 0 && s.queueSize() == 0 && !contracted());
+	assert(s.decisionLevel() == 0 && s.queueSize() == 0);
 	if (ClauseHead::satisfied(s)) {
 		detach(s);
 		return true;
@@ -657,9 +678,14 @@ bool Clause::simplify(Solver& s, bool reinit) {
 	// replace any false lits with sentinels
 	for (Literal* r = j; r != t.second; ++r) { *r = negLit(0); }
 	if (!isSmall()) {
-		uint32 size= std::max((uint32)ClauseHead::HEAD_LITS, static_cast<uint32>(j-head_));
-		data_.local.setSize(size);
+		uint32 size     = std::max((uint32)ClauseHead::HEAD_LITS, static_cast<uint32>(j-head_));
 		data_.local.idx = 0;
+		data_.local.setSize(size);
+		if (j != t.second && learnt() && !strengthened()) {
+			// mark last literal so that we can recompute alloc size later
+			t.second[-1].watch();
+			data_.local.markStrengthened();
+		}
 		if (reinit && size > 3) {
 			detach(s);
 			std::random_shuffle(head_, j, s.strategies().rng);
@@ -675,9 +701,9 @@ bool Clause::simplify(Solver& s, bool reinit) {
 	return j <= t.first && ClauseHead::toImplication(s);
 }
 
-bool Clause::isSatisfied(const Solver& s, LitVec& freeLits) {
-	if (ClauseHead::satisfied(s)) {
-		return true;
+uint32 Clause::isOpen(const Solver& s, const TypeSet& x, LitVec& freeLits) {
+	if (!x.inSet(ClauseHead::type()) || ClauseHead::satisfied(s)) {
+		return 0;
 	}
 	assert(s.queueSize() == 0 && "Watches might be false!");
 	freeLits.push_back(head_[0]);
@@ -691,10 +717,10 @@ bool Clause::isSatisfied(const Solver& s, LitVec& freeLits) {
 		}
 		else if (v == trueValue(*r)) {
 			std::swap(head_[2], *r);
-			return true;
+			return 0;
 		}
 	}
-	return false;
+	return ClauseHead::type();
 }
 
 void Clause::undoLevel(Solver& s) {
@@ -719,9 +745,7 @@ void Clause::undoLevel(Solver& s) {
 void Clause::toLits(LitVec& out) const {
 	out.insert(out.end(), head_, (head_+ClauseHead::HEAD_LITS)-isSentinel(head_[2]));
 	LitRange t = const_cast<Clause&>(*this).tail();
-	if (contracted()) {
-		while (!t.second++->watched()) {;}
-	}
+	if (contracted()) { while (!t.second++->watched()) {;} }
 	out.insert(out.end(), t.first, t.second);
 }
 
@@ -790,6 +814,10 @@ Literal* Clause::removeFromTail(Solver& s, Literal* it, Literal* end) {
 		else          { data_.local.clearContracted(); }
 		end = j;
 	}
+	if (learnt() && !isSmall() && !strengthened()) {
+		end->watch();
+		data_.local.markStrengthened();
+	}
 	return end;
 }
 uint32 Clause::size() const {
@@ -808,14 +836,15 @@ ClauseHead* SharedLitsClause::newClause(Solver& s, SharedLiterals* shared_lits,
 
 SharedLitsClause::SharedLitsClause(Solver& s, SharedLiterals* shared_lits, const Literal* w, const ClauseInfo& e, bool addRef) 
 	: ClauseHead(e) {
-	static_assert(sizeof(SharedLitsClause) <= 32);
+	static_assert(sizeof(SharedLitsClause) <= 32, "Unsupported Padding");
 	data_.shared = addRef ? shared_lits->share() : shared_lits;
 	std::memcpy(head_, w, std::min((uint32)ClauseHead::HEAD_LITS, shared_lits->size())*sizeof(Literal));
 	attach(s);
+	if (learnt()) { s.addLearntBytes(32); }
 }
 
 Constraint* SharedLitsClause::cloneAttach(Solver& other) {
-	return SharedLitsClause::newClause(other, data_.shared, ClauseInfo().setType(this->type()), head_);
+	return SharedLitsClause::newClause(other, data_.shared, ClauseInfo(this->type()), head_);
 }
 
 bool SharedLitsClause::updateWatch(Solver& s, uint32 pos) {
@@ -842,15 +871,14 @@ bool SharedLitsClause::updateWatch(Solver& s, uint32 pos) {
 }
 
 void SharedLitsClause::reason(Solver& s, Literal p, LitVec& out) {
-	ClauseHead::bumpActivity();
 	LitVec::size_type i = out.size();
 	for (const Literal* r = data_.shared->begin(), *end = data_.shared->end(); r != end; ++r) {
 		assert(s.isFalse(*r) || *r == p);
 		if (*r != p) { out.push_back(~*r); }
 	}
-	if (s.strategies().updateLbd && ClauseHead::type() != Constraint_t::static_constraint) {
-		uint32 n = s.computeLbd(p, &out[0]+i, &out[0]+out.size());
-		if (n < info_.data.lbd) { setLbd( n ); }
+	if (learnt()) { 
+		ClauseHead::bumpActivity();
+		setLbd(s.updateLearnt(p, &out[0]+i, &out[0]+out.size(), lbd(), !hasLbd())); 
 	}
 }
 
@@ -893,7 +921,7 @@ bool SharedLitsClause::simplify(Solver& s, bool reinit) {
 		detach(s);
 		SharedLitsClause::destroy(0, false);
 		// construct short clause in "this"
-		ClauseHead* h = Clause::newShortClause(s, this, lits, static_cast<uint32>(j-lits), ClauseInfo());
+		ClauseHead* h = Clause::newClause(this, s, lits, static_cast<uint32>(j-lits), ClauseInfo());
 		// restore extra data - safe because memory layout has not changed!
 		info_.rep = rep;
 		return h->simplify(s, reinit);
@@ -919,13 +947,14 @@ void SharedLitsClause::destroy(Solver* s, bool detachFirst) {
 	this->~SharedLitsClause();
 	if (s) {
 		// return memory to solver
+		s->freeLearntBytes(32);
 		s->freeSmall(mem);
 	}
 }
 
-bool SharedLitsClause::isSatisfied(const Solver& s, LitVec& freeLits) {
-	if (ClauseHead::satisfied(s)) {
-		return true;
+uint32 SharedLitsClause::isOpen(const Solver& s, const TypeSet& x, LitVec& freeLits) {
+	if (!x.inSet(ClauseHead::type()) || ClauseHead::satisfied(s)) {
+		return 0;
 	}
 	Literal* head = head_;
 	ValueRep v;
@@ -935,10 +964,10 @@ bool SharedLitsClause::isSatisfied(const Solver& s, LitVec& freeLits) {
 		}
 		else if (v == trueValue(*r)) {
 			head[2] = *r; // remember as cache literal
-			return true;
+			return 0;
 		}
 	}
-	return false;
+	return ClauseHead::type();
 }
 
 void SharedLitsClause::toLits(LitVec& out) const {
@@ -956,19 +985,23 @@ uint32 SharedLitsClause::size() const { return data_.shared->size(); }
 /////////////////////////////////////////////////////////////////////////////////////////
 // LoopFormula
 /////////////////////////////////////////////////////////////////////////////////////////
-LoopFormula::LoopFormula(Solver& s, uint32 size, Literal* bodyLits, uint32 numBodies, uint32 bodyToWatch, const ClauseInfo& info) {
-	activity_       = info.activity();
-	lbd_            = info.lbd();
-	end_            = numBodies + 2;
-	size_           = end_+1;
-	other_          = end_-1;
-	lits_[0]        = Literal();  // Starting sentinel
-	lits_[end_-1]   = Literal();  // Position of active atom
-	lits_[end_]     = Literal();  // Ending sentinel - active part
+LoopFormula* LoopFormula::newLoopFormula(Solver& s, Literal* bodyLits, uint32 numBodies, uint32 bodyToWatch, uint32 numAtoms, const Activity& act) {
+	uint32 bytes = sizeof(LoopFormula) + (numBodies+numAtoms+3) * sizeof(Literal);
+	void* mem    = Detail::alloc( bytes );
+	s.addLearntBytes(bytes);
+	return new (mem) LoopFormula(s, numBodies+numAtoms, bodyLits, numBodies, bodyToWatch, act);
+}
+LoopFormula::LoopFormula(Solver& s, uint32 size, Literal* bodyLits, uint32 numBodies, uint32 bodyToWatch, const Activity& act)
+	: act_(act) {
+	end_          = numBodies + 2;
+	size_         = end_+1;
+	other_        = end_-1;
+	lits_[0]      = Literal();  // Starting sentinel
+	lits_[end_-1] = Literal();  // Position of active atom
+	lits_[end_]   = Literal();  // Ending sentinel - active part
 	for (uint32 i = size_; i != size+3; ++i) {
 		lits_[i] = Literal();
 	}
-
 	// copy bodies: S B1...Bn, watch one
 	std::memcpy(lits_+1, bodyLits, numBodies * sizeof(Literal));
 	s.addWatch(~lits_[1+bodyToWatch], this, ((1+bodyToWatch)<<1)+1);
@@ -976,29 +1009,35 @@ LoopFormula::LoopFormula(Solver& s, uint32 size, Literal* bodyLits, uint32 numBo
 }
 
 void LoopFormula::destroy(Solver* s, bool detach) {
-	if (s && detach) {
-		for (uint32 x = 1; x != end_-1; ++x) {
-			if (lits_[x].watched()) {
-				s->removeWatch(~lits_[x], this);
-				lits_[x].clearWatch();
+	if (s) {
+		if (detach) {
+			for (uint32 x = 1; x != end_-1; ++x) {
+				if (lits_[x].watched()) {
+					s->removeWatch(~lits_[x], this);
+					lits_[x].clearWatch();
+				}
+			}
+			if (lits_[end_-1].watched()) {
+				lits_[end_-1].clearWatch();
+				for (uint32 x = end_+1; x != size_; ++x) {
+					s->removeWatch(~lits_[x], this);
+					lits_[x].clearWatch();
+				}
 			}
 		}
-		if (lits_[end_-1].watched()) {
-			lits_[end_-1].clearWatch();
-			for (uint32 x = end_+1; x != size_; ++x) {
-				s->removeWatch(~lits_[x], this);
-				lits_[x].clearWatch();
-			}
+		if (lits_[0].watched()) {
+			while (lits_[size_++].asUint() != 3u) { ; }
 		}
+		s->freeLearntBytes(sizeof(LoopFormula) + (size_ * sizeof(Literal)));
 	}
 	void* mem = static_cast<Constraint*>(this);
 	this->~LoopFormula();
-	::operator delete(mem);
+	Detail::free(mem);
 }
 
 
 void LoopFormula::addAtom(Literal atom, Solver& s) {
-	++activity_;
+	act_.bumpAct();
 	uint32 pos = size_++;
 	assert(isSentinel(lits_[pos]));
 	lits_[pos] = atom;
@@ -1013,7 +1052,7 @@ void LoopFormula::updateHeuristic(Solver& s) {
 	Literal saved = lits_[end_-1];
 	for (uint32 x = end_+1; x != size_; ++x) {
 		lits_[end_-1] = lits_[x];
-		s.strategies().heuristic->newConstraint(s, lits_+1, end_-1, Constraint_t::learnt_loop);
+		s.heuristic()->newConstraint(s, lits_+1, end_-1, Constraint_t::learnt_loop);
 	}
 	lits_[end_-1] = saved;
 }
@@ -1115,15 +1154,12 @@ void LoopFormula::reason(Solver& s, Literal p, LitVec& lits) {
 	if (other_ != end_-1) {
 		lits.push_back(~lits_[end_-1]);
 	}
-	++activity_;
-	if (s.strategies().updateLbd) {
-		uint32 n = s.computeLbd(p, &lits[0]+os, &lits[0]+lits.size());
-		if (n < lbd_) { lbd_ = n; }
-	}
+	act_.setLbd(s.updateLearnt(p, &lits[0]+os, &lits[0]+lits.size(), act_.lbd()));
+	act_.bumpAct();
 }
 
 bool LoopFormula::minimize(Solver& s, Literal p, CCMinRecursive* rec) {
-	++activity_;
+	act_.bumpAct();
 	for (uint32 x = 1; x != end_-1; ++x) {
 		if (lits_[x] != p && !s.ccMinimize(~lits_[x], rec)) {
 			return false;
@@ -1149,24 +1185,22 @@ bool LoopFormula::locked(const Solver& s) const {
 	return false;
 }
 
-bool LoopFormula::isSatisfied(const Solver& s, LitVec& freeLits) {
-	if (other_ != end_-1 && s.isTrue(lits_[other_])) return true;
+uint32 LoopFormula::isOpen(const Solver& s, const TypeSet& x, LitVec& freeLits) {
+	if (!x.inSet(Constraint_t::learnt_loop) || (other_ != end_-1 && s.isTrue(lits_[other_]))) {
+		return 0;
+	}
 	for (uint32 x = 1; x != end_-1; ++x) {
 		if (s.isTrue(lits_[x])) {
 			other_ = x;
-			return true;
+			return 0;
 		}
 		else if (!s.isFalse(lits_[x])) { freeLits.push_back(lits_[x]); }
 	}
-	bool sat = true;
 	for (uint32 x = end_+1; x != size_; ++x) {
-		if (s.value(lits_[x].var()) == value_free) {
-			freeLits.push_back(lits_[x]);
-			sat = false;
-		}
-		else sat &= s.isTrue(lits_[x]);
+		if      (s.value(lits_[x].var()) == value_free) { freeLits.push_back(lits_[x]); }
+		else if (s.isTrue(lits_[x]))                    { return 0; }
 	}
-	return sat;
+	return Constraint_t::learnt_loop;
 }
 
 bool LoopFormula::simplify(Solver& s, bool) {
@@ -1216,6 +1250,10 @@ bool LoopFormula::simplify(Solver& s, bool) {
 			lits_[i].clearWatch();
 		}
 	}
+	if (j != size_ && !lits_[0].watched()) {
+		lits_[size_-1].asUint() = 3u;
+		lits_[0].watch();
+	}
 	size_         = j;
 	end_          = newEnd;
 	lits_[end_]   = Literal();
