@@ -19,6 +19,7 @@
 //
 #include <clasp/heuristics.h>
 #include <algorithm>
+#include <limits>
 namespace Clasp {
 /////////////////////////////////////////////////////////////////////////////////////////
 // Lookback selection strategies
@@ -48,15 +49,14 @@ uint32 momsScore(const Solver& s, Var v) {
 #define BERK_MAX_MOMS_DECS 50
 #define BERK_MAX_DECAY 65534
 
-ClaspBerkmin::ClaspBerkmin(uint32 maxB, bool loops, bool moms, bool huang, bool scoreSet) 
-	: order_(huang, moms, loops)
+ClaspBerkmin::ClaspBerkmin(uint32 maxB) 
+	: order_(false)
 	, topConflict_(UINT32_MAX)
-	, topLoop_(UINT32_MAX)
+	, topOther_(UINT32_MAX)
 	, front_(1) 
 	, cacheSize_(5)
 	, numVsids_(0)
-	, maxBerkmin_(maxB == 0 ? UINT32_MAX : maxB)
-	, scoreMulti_(!scoreSet) {
+	, maxBerkmin_(maxB == 0 ? UINT32_MAX : maxB) {
 }
 
 Var ClaspBerkmin::getMostActiveFreeVar(const Solver& s) {
@@ -116,12 +116,13 @@ Var ClaspBerkmin::getTopMoms(const Solver& s) {
 }
 
 void ClaspBerkmin::Order::init(const Solver& s) {
-	if (reinit) { score.clear(); }
+	if (s.strategies().heuReinit) { score.clear(); }
 	score.resize(s.numVars()+1);
 }
 
 void ClaspBerkmin::startInit(const Solver& s) {
 	if (order_.score.empty()) {
+		order_.huang = s.strategies().berkHuang != 0;
 		rng_.srand(s.strategies().rng.seed());
 	}
 	order_.init(s);
@@ -132,8 +133,8 @@ void ClaspBerkmin::startInit(const Solver& s) {
 	cacheFront_ = cache_.end();
 	
 	freeLits_.clear();
-	freeLoopLits_.clear();
-	topConflict_ = topLoop_ = (uint32)-1;
+	freeOtherLits_.clear();
+	topConflict_ = topOther_ = (uint32)-1;
 	
 	front_ = 1;
 	order_.decay = numVsids_ = 0;
@@ -141,7 +142,7 @@ void ClaspBerkmin::startInit(const Solver& s) {
 
 void ClaspBerkmin::endInit(Solver& s) {
 	if (initHuang()) {
-		const bool clearScore = order_.moms;
+		const bool clearScore = s.strategies().heuMoms != 0;
 		// initialize preferred values of vars
 		cache_.clear();
 		for (Var v = 1; v <= s.numVars(); ++v) {
@@ -154,7 +155,7 @@ void ClaspBerkmin::endInit(Solver& s) {
 		}
 		initHuang(false);
 	}
-	if (!order_.moms || s.numFreeVars() > BERK_MAX_MOMS_VARS) {
+	if (s.strategies().heuMoms == 0 || s.numFreeVars() > BERK_MAX_MOMS_VARS) {
 		hasActivities(true);
 	}
 	std::stable_sort(cache_.begin(), cache_.end(), Order::Compare(&order_));
@@ -173,9 +174,9 @@ void ClaspBerkmin::newConstraint(const Solver&, const Literal* first, LitVec::si
 }
 
 void ClaspBerkmin::updateReason(const Solver& s, const LitVec& lits, Literal r) {
-	const bool sc = scoreMulti_;
+	const bool ms = s.strategies().berkOnce == 0;
 	for (LitVec::size_type i = 0, e = lits.size(); i != e; ++i) {
-		if (sc || !s.seen(lits[i])) {
+		if (ms || !s.seen(lits[i])) {
 			order_.inc(~lits[i]);
 		}
 	}
@@ -185,14 +186,13 @@ void ClaspBerkmin::updateReason(const Solver& s, const LitVec& lits, Literal r)
 bool ClaspBerkmin::bump(const Solver&, const WeightLitVec& lits, double adj) {
 	for (WeightLitVec::const_iterator it = lits.begin(), end = lits.end(); it != end; ++it) {
 		uint32 xf = order_.decayedScore(it->first.var()) + static_cast<weight_t>(it->second*adj);
-		while (xf >= UINT32_MAX/2) { xf >>= 1; }
-		order_.score[it->first.var()].act = (uint16)xf;
+		order_.score[it->first.var()].act = (uint16)std::min(xf, UINT32_MAX>>16);
 	}
 	return true;
 }
 
 void ClaspBerkmin::undoUntil(const Solver&, LitVec::size_type) {
-	topConflict_ = topLoop_ = static_cast<uint32>(-1);
+	topConflict_ = topOther_ = static_cast<uint32>(-1);
 	front_ = 1;
 	cache_.clear();
 	cacheFront_ = cache_.end();
@@ -202,44 +202,42 @@ void ClaspBerkmin::undoUntil(const Solver&, LitVec::size_type) {
 	numVsids_ = 0;
 }
 
-bool ClaspBerkmin::hasTopUnsat(Solver& s, uint32& maxIdx, uint32 minIdx, ConstraintType t) {
-	while (maxIdx != minIdx) {
-		LearntConstraint& c = s.getLearnt(maxIdx-1);
-		if (c.type() == t && !c.isSatisfied(s, freeLits_)) {
-			return true;
-		}
-		--maxIdx;
-		freeLits_.clear();
-	}
-	return false;
-}
-
 bool ClaspBerkmin::hasTopUnsat(Solver& s) {
 	topConflict_  = std::min(s.numLearntConstraints(), topConflict_);
-	topLoop_      = std::min(s.numLearntConstraints(), topLoop_);
-	freeLoopLits_.clear();
+	topOther_     = std::min(s.numLearntConstraints(), topOther_);
+	assert(topConflict_ <= topOther_);
+	freeOtherLits_.clear();
 	freeLits_.clear();
-	if (topConflict_ > topLoop_ && hasTopUnsat(s, topConflict_, topLoop_, Constraint_t::learnt_conflict)) {
-		return true;
-	}
-	if (order_.loops && topLoop_ > topConflict_ && hasTopUnsat(s, topLoop_, topConflict_, Constraint_t::learnt_loop)) {
-		freeLits_.swap(freeLoopLits_);
+	TypeSet ts;
+	if (s.strategies().heuOther != 0) {
+		ts.addSet(Constraint_t::learnt_loop);
+		if (s.strategies().heuOther == 2) { ts.addSet(Constraint_t::learnt_other); }
+		while (topOther_ > topConflict_) {
+			if (s.getLearnt(topOther_-1).isOpen(s, ts, freeLits_) != 0) {
+				freeLits_.swap(freeOtherLits_);
+				ts.m = 0;
+				break;
+			}
+			--topOther_;
+			freeLits_.clear();
+		}
 	}
+	ts.addSet(Constraint_t::learnt_conflict);
 	uint32 stopAt = topConflict_ < maxBerkmin_ ? 0 : topConflict_ - maxBerkmin_;
 	while (topConflict_ != stopAt) {
-		LearntConstraint& c = s.getLearnt(topConflict_-1);
-		if (c.type() == Constraint_t::learnt_conflict && !c.isSatisfied(s, freeLits_)) {
-			break;
-		}
-		else if (order_.loops && freeLoopLits_.empty() && c.type() == Constraint_t::learnt_loop && !c.isSatisfied(s, freeLits_)) {
-			topLoop_  = topConflict_;
-			freeLits_.swap(freeLoopLits_);
+		uint32 x = s.getLearnt(topConflict_-1).isOpen(s, ts, freeLits_);
+		if (x != 0) {
+			if (x == Constraint_t::learnt_conflict) { break; }
+			topOther_  = topConflict_;
+			freeLits_.swap(freeOtherLits_);
+			ts.m = 0;
+			ts.addSet(Constraint_t::learnt_conflict);
 		}
 		--topConflict_;
 		freeLits_.clear();
 	}
-	if (freeLoopLits_.empty())  topLoop_ = topConflict_;
-	if (freeLits_.empty())      freeLoopLits_.swap(freeLits_);
+	if (freeOtherLits_.empty())  topOther_ = topConflict_;
+	if (freeLits_.empty())      freeOtherLits_.swap(freeLits_);
 	return !freeLits_.empty();
 }
 
@@ -293,24 +291,27 @@ Literal ClaspBerkmin::selectRange(Solver& s, const Literal* first, const Literal
 	return c == 1 ? candidates[0] : candidates[rng_.irand(c)];
 }
 
-Literal ClaspBerkmin::selectLiteral(Solver& s, Var var, bool vsids) {
-	Literal l;
-	if ( (l = savedLiteral(s,var)) == posLit(0) ) {
-		int32 w0 = vsids ? (int32)s.estimateBCP(posLit(var), 5) : order_.occ(var);
-		int32 w1 = vsids ? (int32)s.estimateBCP(negLit(var), 5) : 0;
+Literal ClaspBerkmin::selectLiteral(Solver& s, Var v, bool vsids) const {
+	ValueRep pv = s.prefValue(v);
+	ValueRep sv = s.savedValue(v);
+	if (sv == 0 && pv == 0) {
+		int32 w0 = vsids ? (int32)s.estimateBCP(posLit(v), 5) : order_.occ(v);
+		int32 w1 = vsids ? (int32)s.estimateBCP(negLit(v), 5) : 0;
 		if (w1 == 1 && w0 == w1) {
 			// no binary bcp - use occurrences
-			w0 = order_.occ(var);
+			w0 = order_.occ(v);
 			w1 = 0;
 		}
-		return w0 != w1 
-			? Literal(var, (w0-w1)<0)
-			: s.sharedContext()->preferredLiteralByType(var);
+		return w0 != w1 ? Literal(v, (w0-w1)<0) : s.defaultLiteral(v);
 	}
-	else if (order_.huang && (order_.occ(var)*(-1+(2*l.sign()))) > 32) {
-		l = ~l;
+	else if (pv == 0) { 
+		Literal x(v, valSign(sv));
+		if (order_.huang && (order_.occ(v)*(-1+(2*x.sign()))) > 32) {
+			x = ~x;
+		}
+		return x;
 	}
-	return l;
+	else { return Literal(v, valSign(pv)); }
 }
 
 void ClaspBerkmin::Order::resetDecay() {
@@ -323,18 +324,19 @@ void ClaspBerkmin::Order::resetDecay() {
 /////////////////////////////////////////////////////////////////////////////////////////
 // ClaspVmtf selection strategy
 /////////////////////////////////////////////////////////////////////////////////////////
-ClaspVmtf::ClaspVmtf(LitVec::size_type mtf, bool loops) : MOVE_TO_FRONT(mtf), loops_(loops), reinit_(true) { 
+ClaspVmtf::ClaspVmtf(LitVec::size_type mtf) : MOVE_TO_FRONT(mtf >= 2 ? mtf : 2) { 
 }
 
 
 void ClaspVmtf::startInit(const Solver& s) {
-	if (reinit_) { score_.clear(); vars_.clear(); }
+	if (s.strategies().heuReinit) { score_.clear(); vars_.clear(); }
 	score_.resize(s.numVars()+1, VarInfo(vars_.end()));
 }
 
 void ClaspVmtf::endInit(Solver& s) {
-	if (reinit_) vars_.clear();
-	bool moms = reinit_ || score_[0].activity_ == 0;
+	if (s.strategies().heuReinit)     { vars_.clear(); }
+	if (s.strategies().heuOther == 3) { s.strategies().heuOther = 0; }
+	bool moms = s.strategies().heuMoms != 0;
 	for (Var v = 1; v <= s.numVars(); ++v) {
 		if (s.value(v) == value_free && score_[v].pos_ == vars_.end()) {
 			if (moms) {
@@ -367,10 +369,11 @@ void ClaspVmtf::newConstraint(const Solver& s, const Literal* first, LitVec::siz
 	if (t != Constraint_t::static_constraint) {
 		LessLevel comp(s, score_);
 		VarVec::size_type maxMove = t == Constraint_t::learnt_conflict ? MOVE_TO_FRONT : MOVE_TO_FRONT/2;
+		const bool upAct = t == Constraint_t::learnt_conflict || s.strategies().heuOther >= uint32(t-1);
 		for (LitVec::size_type i = 0; i < size; ++i, ++first) {
 			Var v = first->var(); 
 			score_[v].occ_ += 1 - (((int)first->sign())<<1);
-			if (t == Constraint_t::learnt_conflict || loops_) {
+			if (upAct) {
 				++score_[v].activity(decay_);
 				if (mtf_.size() < maxMove) {
 					mtf_.push_back(v);
@@ -395,16 +398,6 @@ void ClaspVmtf::newConstraint(const Solver& s, const Literal* first, LitVec::siz
 	} 
 }
 
-Literal ClaspVmtf::getLiteral(const Solver& s, Var v) const {
-	Literal r;
-	if ( (r = savedLiteral(s, v)) == posLit(0) ) {
-		r = score_[v].occ_== 0
-			? s.sharedContext()->preferredLiteralByType(v)
-			: Literal(v, score_[v].occ_ < 0 );
-	}
-	return r;
-}
-
 void ClaspVmtf::undoUntil(const Solver&, LitVec::size_type) {
 	front_ = vars_.begin();
 }
@@ -432,11 +425,11 @@ Literal ClaspVmtf::doSelect(Solver& s) {
 			++distance;
 		} while (s.value(*v2) != value_free);
 		c = (score_[*front_].activity(decay_) + (distance<<1)+ 3) > score_[*v2].activity(decay_)
-				? getLiteral(s, *front_)
-				: getLiteral(s, *v2);
+		    ? selectLiteral(s, *front_, score_[*front_].occ_)
+		    : selectLiteral(s, *v2, score_[*v2].occ_);
 	}
 	else {
-		c = getLiteral(s, *front_);
+		c = selectLiteral(s, *front_, score_[*front_].occ_);
 	}
 	return c;
 }
@@ -454,24 +447,24 @@ Literal ClaspVmtf::selectRange(Solver&, const Literal* first, const Literal* las
 /////////////////////////////////////////////////////////////////////////////////////////
 // ClaspVsids selection strategy
 /////////////////////////////////////////////////////////////////////////////////////////
-ClaspVsids::ClaspVsids(double decay, bool loops) 
+ClaspVsids::ClaspVsids(double decay) 
 	: vars_(GreaterActivity(score_)) 
-	, decay_(1.0 / decay)
-	, inc_(1.0)
-	, scoreLoops_(loops) 
-	, reinit_(true) {}
+	, decay_(1.0 / std::max(0.01, std::min(1.0, decay)))
+	, inc_(1.0) {}
 
 void ClaspVsids::startInit(const Solver& s) {
-	if (reinit_)score_.clear();
+	if (s.strategies().heuReinit) score_.clear();
 	score_.resize(s.numVars()+1);
 }
+
 void ClaspVsids::endInit(Solver& s) {
-	if (reinit_) vars_.clear();
-	bool moms = vars_.empty();
+	if (s.strategies().heuReinit)     { vars_.clear(); }
+	if (s.strategies().heuOther == 3) { s.strategies().heuOther = 0; }
+	bool moms = s.strategies().heuMoms != 0;
 	inc_ = 1.0;
 	double maxS = 0.0;
 	for (Var v = 1; v <= s.numVars(); ++v) {
-		if (s.value(v) == value_free && (reinit_ || !vars_.is_in_queue(v))) {
+		if (s.value(v) == value_free && !vars_.is_in_queue(v)) {
 			if (moms) {
 				// initialize activity to moms-score
 				score_[v].first = momsScore(s, v);
@@ -495,11 +488,29 @@ void ClaspVsids::simplify(const Solver& s, LitVec::size_type i) {
 	}
 }
 
-void ClaspVsids::newConstraint(const Solver&, const Literal* first, LitVec::size_type size, ConstraintType t) {
+void ClaspVsids::normalize() {
+	const double min  = std::numeric_limits<double>::min();
+	const double minD = min * 1e100;
+	inc_             *= 1e-100;
+	for (LitVec::size_type i = 0; i != score_.size(); ++i) {
+		double d = score_[i].first;
+		if (d > 0) {
+			// keep relative ordering but 
+			// actively avoid denormals
+			d += minD;
+			d *= 1e-100;
+		}
+		score_[i].first = d;
+	}
+}
+
+void ClaspVsids::newConstraint(const Solver& s, const Literal* first, LitVec::size_type size, ConstraintType t) {
 	if (t != Constraint_t::static_constraint) {
+		const uint32 other = s.strategies().heuOther != 3 ? s.strategies().heuOther : 0;
+		const bool upAct   = t == Constraint_t::learnt_conflict || (other >= uint32(t-1));
 		for (LitVec::size_type i = 0; i < size; ++i, ++first) {
 			score_[first->var()].second += 1 - (first->sign() + first->sign());
-			if (t == Constraint_t::learnt_conflict || scoreLoops_) {
+			if (upAct) {
 				updateVarActivity(first->var());
 			}
 		}
@@ -530,18 +541,10 @@ void ClaspVsids::undoUntil(const Solver& s , LitVec::size_type st) {
 }
 
 Literal ClaspVsids::doSelect(Solver& s) {
-	Var var;
 	while ( s.value(vars_.top()) != value_free ) {
 		vars_.pop();
 	}
-	var = vars_.top();
-	Literal r;
-	if ( (r = savedLiteral(s, var)) == posLit(0) ) {
-		r = score_[var].second == 0
-			? s.sharedContext()->preferredLiteralByType(var)
-			: Literal( var, score_[var].second < 0 );
-	}
-	return r;
+	return selectLiteral(s, vars_.top(), score_[vars_.top()].second);
 }
 
 Literal ClaspVsids::selectRange(Solver&, const Literal* first, const Literal* last) {
