@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2010, Benjamin Kaufmann
+// Copyright (c) 2010-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -50,7 +50,7 @@ bool MinimizeConstraint::attach(Solver& s) {
 	uint32 numL = 0;
 	VarVec up;
 	const SharedData* d = shared_;
-	const bool heuristic= (s.strategies().heuOpts & SolverStrategies::heu_use_static_sign) != 0;
+	const bool heuristic= (s.strategies().optHeu & SolverStrategies::opt_sign) != 0;
 	for (const WeightLiteral* it = d->lits; !isSentinel(it->first); ++it, ++numL) {
 		if (s.value(it->first.var()) == value_free) {
 			s.addWatch(it->first, this, numL);
@@ -286,7 +286,7 @@ bool MinimizeConstraint::integrateNext(Solver& s) {
 }
 
 bool MinimizeConstraint::modelHeuristic(Solver& s) {
-	const bool modelHeu  = (s.strategies().heuOpts & SolverStrategies::heu_use_model) != 0;
+	const bool modelHeu  = (s.strategies().optHeu & SolverStrategies::opt_model) != 0;
 	const bool heuristic = modelHeu || (s.queueSize() == 0 && s.decisionLevel() == s.rootLevel());
 	if (heuristic && s.propagate()) {
 		for (const WeightLiteral* w = pos_; !isSentinel(w->first); ++w) {
@@ -738,7 +738,8 @@ bool MinimizeBuilder::CmpByWeight::operator()(const LitRep& lhs, const LitRep& r
 		wLhs = wLhs->next;
 		wRhs = wRhs->next;
 	}
-	return wLhs != 0;
+	return (wLhs && wLhs->weight > 0)
+	  ||   (wRhs && wRhs->weight < 0);
 }
 int MinimizeBuilder::CmpByWeight::compare(const LitRep& lhs, const LitRep& rhs) const {
 	if (this->operator()(lhs, rhs)) return 1;
