@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2010, Benjamin Kaufmann
+// Copyright (c) 2010-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -17,26 +17,16 @@
 // along with Clasp; if not, write to the Free Software
 // Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 //
-#ifndef DISABLE_MULTI_THREADING
+#if WITH_THREADS
 #include <clasp/parallel_solve.h>
 #include <clasp/solver.h>
 #include <clasp/clause.h>
 #include <clasp/clasp_facade.h>
 #include <clasp/util/timer.h>
 #include <clasp/minimize_constraint.h>
+#include <clasp/util/mutex.h>
 #include <tbb/concurrent_queue.h>
 
-#if !defined(NDEBUG) && !defined(ENABLE_LOG)
-#define ENABLE_LOG 1
-#endif
-
-#if defined(ENABLE_LOG) && ENABLE_LOG == 1
-#include <stdio.h>
-#define LOG(id, fmt, ...) ( fprintf(stderr, "THREAD[%u]: " fmt "\n", (id), ##__VA_ARGS__), fflush(stderr) )
-#else
-#define LOG(id, fmt, ...)
-#endif
-
 namespace Clasp { namespace mt {
 /////////////////////////////////////////////////////////////////////////////////////////
 // BarrierSemaphore
@@ -54,25 +44,25 @@ public:
 		active_  = maxParties;
 	}
 	// Returns the current semaphore counter.
-	int  counter()   { std::unique_lock<tbb::mutex> lock(semMutex_); return counter_; }
+	int  counter()   { std::lock_guard<std::mutex> lock(semMutex_); return counter_; }
 	// Returns the number of parties required to trip this barrier.
-	int  parties()   { std::unique_lock<tbb::mutex> lock(semMutex_); return active_;  } 
+	int  parties()   { std::lock_guard<std::mutex> lock(semMutex_); return active_;  } 
 	// Returns true if all parties are waiting at the barrier
-	bool active()    { std::unique_lock<tbb::mutex> lock(semMutex_); return unsafe_active(); }
+	bool active()    { std::lock_guard<std::mutex> lock(semMutex_); return unsafe_active(); }
 	
 	// barrier interface
 	
 	// Increases the barrier count, i.e. the number of 
 	// parties required to trip this barrier.
 	void addParty() {
-		std::unique_lock<tbb::mutex> lock(semMutex_);
+		std::lock_guard<std::mutex> lock(semMutex_);
 		++active_;
 	}
 	// Decreases the barrier count and resets the barrier
 	// if reset is true. 
 	// PRE: the thread does not itself wait on the barrier
 	void removeParty(bool reset) {
-		std::unique_lock<tbb::mutex> lock(semMutex_);
+		std::unique_lock<std::mutex> lock(semMutex_);
 		assert(active_ > 0);
 		--active_;
 		if      (reset)           { unsafe_reset(0); }
@@ -84,13 +74,13 @@ public:
 	// Applications shall use this value to designate one thread as the
 	// leader that will eventually reset the barrier thereby unblocking the other threads.
 	bool wait() {
-		std::unique_lock<tbb::mutex> lock(semMutex_);
+		std::unique_lock<std::mutex> lock(semMutex_);
 		if (--counter_ >= 0) { counter_ = -1; }
 		return unsafe_wait(lock);
 	}
 	// Resets the barrier and unblocks any waiting threads.
 	void reset(uint32 semCount = 0) {
-		std::unique_lock<tbb::mutex> lock(semMutex_);
+		std::lock_guard<std::mutex> lock(semMutex_);
 		unsafe_reset(semCount);
 	}
 	// semaphore interface
@@ -101,7 +91,7 @@ public:
 	// Returns false to signal that all but the calling thread
 	// are currently blocked.
 	bool down() {
-		std::unique_lock<tbb::mutex> lock(semMutex_);
+		std::unique_lock<std::mutex> lock(semMutex_);
 		if (--counter_ >= 0) { return true; }
 		return !unsafe_wait(lock);
 	}
@@ -111,7 +101,7 @@ public:
 	void up() {
 		bool notify;
 		{
-			std::unique_lock<tbb::mutex> lock(semMutex_);
+			std::lock_guard<std::mutex> lock(semMutex_);
 			notify    = (++counter_ < 1);
 		}
 		if (notify) semCond_.notify_one();
@@ -127,7 +117,7 @@ private:
 		if (prev < 0) { semCond_.notify_all(); }
 	}
 	// Returns true for the leader, else false
-	bool    unsafe_wait(std::unique_lock<tbb::mutex>& lock) {
+	bool    unsafe_wait(std::unique_lock<std::mutex>& lock) {
 		assert(counter_ < 0);
 		// don't put the last thread to sleep!
 		if (!unsafe_active()) {
@@ -136,7 +126,7 @@ private:
 		return unsafe_active();
 	}	
 	cv         semCond_;  // waiting threads
-	tbb::mutex semMutex_; // mutex for updating counter
+	std::mutex semMutex_; // mutex for updating counter
 	int        counter_;  // semaphore's counter
 	int        active_;   // number of active threads
 };
@@ -161,12 +151,12 @@ struct ParallelSolve::SharedData {
 		msg_sync_unsat     = (sync_flag | unsat_flag | restart_flag),
 		msg_split          = split_flag
 	};
-	SharedData() : path(0) { reset(0, 1); }
-	void reset(SharedContext* a_ctx, uint32 numT) {
+	SharedData() : path(0) { reset(0); }
+	void reset(SharedContext* a_ctx) {
 		clearQueue();
 		syncT.reset();
-		workSem.unsafe_init(0, numT);
-		globalR.reset();
+		workSem.unsafe_init(0, a_ctx ? a_ctx->numSolvers() : 0);
+		globalR     = ScheduleStrategy::none();
 		maxConflict = globalR.current();
 		ctx         = a_ctx;
 		path        = 0;
@@ -203,7 +193,7 @@ struct ParallelSolve::SharedData {
 	SharedContext*      ctx;         // shared context object
 	const LitVec*       path;        // initial guiding path - typically empty
 	Timer<RealTime>     syncT;       // thread sync time
-	tbb::mutex          modelM;      // model-mutex 
+	std::mutex          modelM;      // model-mutex 
 	BarrierSemaphore    workSem;     // work-semaphore
 	queue               workQ;       // work-queue
 	uint32              nextId;      // next solver id to use
@@ -238,244 +228,193 @@ void ParallelSolve::SharedData::updateSplitFlag() {
 		else        fetch_and_and(control, ~uint32(split_flag));
 	}
 }
-
-// Callable function object to be used as "main" function
-// for threads. Simply forwards to ParallelSolve::doSolve()
-struct ParallelSolve::EntryPoint {
-	explicit EntryPoint(ParallelSolve* self) : self_(self) {}
-	void operator()(Solver* s, const SolveParams* p) { self_->doSolve(*s, *p, LitVec()); }
-	ParallelSolve* self_;
-};
 /////////////////////////////////////////////////////////////////////////////////////////
 // ParallelSolve
 /////////////////////////////////////////////////////////////////////////////////////////
-ParallelSolve::ParallelSolve()
-	: SolveAlgorithm()
+ParallelSolve::ParallelSolve(SharedContext& ctx, const ParallelSolveOptions& opts)
+	: SolveAlgorithm(opts.limit)
 	, shared_(new SharedData)
 	, thread_(0)
-	, maxThreads_(0)
 	, maxRestarts_(0)
 	, intGrace_(1024)
 	, intFlags_(ClauseCreator::clause_not_root_sat | ClauseCreator::clause_no_add)
-	, intHeuristic_(false)
-	, forceGP_(true) {
+	, error_(0)
+	, initialGp_(opts.mode == ParallelSolveOptions::mode_split ? gp_split : gp_fixed) {
+	assert(ctx.numSolvers() && "Illegal number of threads");
+	shared_->reset(&ctx);
+	shared_->setControl(opts.mode == ParallelSolveOptions::mode_split ? SharedData::allow_gp_flag : SharedData::forbid_restart_flag);
+	shared_->setControl(SharedData::sync_flag); // force initial sync with all threads
+	setRestarts(opts.restarts.maxR, opts.restarts.sched);
+	setIntegrate(opts.integrate.grace, opts.integrate.filter);
+	if (ctx.distribution()) {
+		ctx.setDistributor(new mt::GlobalQueue(ctx.numSolvers(), opts.integrate.topo));
+	}
+	ctx.master()->stats.enableParallelStats();
 }
 
 ParallelSolve::~ParallelSolve() {
-	if (maxThreads_ > 1) {
+	if (shared_->nextId > 1) {
 		// algorithm was not started but there may be active threads -
 		// force orderly shutdown
 		ParallelSolve::terminate();
 		shared_->workSem.removeParty(true);
 		joinThreads();
-		assert(numThreads() == 0);
 	}
-	destroyThread(0);
+	destroyThread(masterId);
 	delete shared_;
 }
 
-uint32 ParallelSolve::hardwareThreads() {
-	return std::thread::hardware_concurrency();
-}
-
-void ParallelSolve::setForceGP(bool gp) {
-	forceGP_ = gp;
-}
-
 void ParallelSolve::setIntegrate(uint32 grace, uint8 filter) {
-	typedef ThreadOptions::Distribution Dist;
+	typedef ParallelSolveOptions::Integration Dist;
 	intGrace_     = grace;
 	intFlags_     = ClauseCreator::clause_no_add;
-	intHeuristic_ = filter == Dist::filter_heuristic;
-	if (filter != Dist::filter_no) { intFlags_ |= ClauseCreator::clause_not_root_sat; }
-	if (filter == Dist::filter_sat){ intFlags_ |= ClauseCreator::clause_not_sat; }
+	if (filter == Dist::filter_heuristic) { store_set_bit(intFlags_, 31); }
+	if (filter != Dist::filter_no)        { intFlags_ |= ClauseCreator::clause_not_root_sat; }
+	if (filter == Dist::filter_sat)       { intFlags_ |= ClauseCreator::clause_not_sat; }
 }
 
 void ParallelSolve::setRestarts(uint32 maxR, const ScheduleStrategy& rs) {
-	shared_->globalR     = rs;
-	shared_->maxConflict = shared_->globalR.current();
 	maxRestarts_         = maxR;
+	shared_->globalR     = maxR ? rs : ScheduleStrategy::none();
+	shared_->maxConflict = shared_->globalR.current();
 }
 
-// prepare parallel solving
-SharedContext* ParallelSolve::init(int numT, SharedContext& ctx) {
-	assert(numT >= 1 && "Illegal number of threads");
-	shared_->reset(&ctx, static_cast<uint32>(numT));
-	reserveThreads();
-	return &ctx;
-}
-
-uint32 ParallelSolve::numThreads() const {
-	return shared_->workSem.parties();
-}
-
-void ParallelSolve::reserveThreads() {
-	destroyThread(0);
-	uint32 numT = numThreads();
-	thread_     = new ParallelHandler*[numT];
-	maxThreads_ = 0;
-	for (uint32 i = 0; i != numT; ++i) {
-#pragma message TODO("replace with CACHE_LINE_ALIGNED alloc")
-		uint32 b  = ((sizeof(ParallelHandler)+63) / 64) * 64;
-		thread_[i]= new (::operator new( b )) ParallelHandler(*this);
-		++maxThreads_;
+void ParallelSolve::addSolver(Solver& s, const SolveParams& p) {
+	if (&s != shared_->ctx->master()) {
+		uint32 id = shared_->nextId++;
+		assert(id < shared_->ctx->numSolvers() && id != masterId);
+		s.setId(id);
+		s.stats.enableStats(shared_->ctx->master()->stats);
+		allocThread(id, s, p);
+		// start in new thread
+		std::thread x(std::mem_fun(&ParallelSolve::solveParallel), this, id);
+		thread_[id]->setThread(x);
 	}
 }
 
+uint32 ParallelSolve::numThreads() const { return shared_->workSem.parties(); }
+
+void ParallelSolve::allocThread(uint32 id, Solver& s, const SolveParams& p) {
+	if (!thread_) {
+		uint32 n = numThreads();
+		thread_  = new ParallelHandler*[n];
+		std::fill(thread_, thread_+n, static_cast<ParallelHandler*>(0));
+	}
+	#pragma message TODO("replace with CACHE_LINE_ALIGNED alloc")
+	uint32 b   = ((sizeof(ParallelHandler)+63) / 64) * 64;
+	thread_[id]= new (::operator new( b )) ParallelHandler(*this, s, p);
+}
+
 void ParallelSolve::destroyThread(uint32 id) {
-	assert(id <= maxThreads_);
-	if (id < maxThreads_) {
+	if (thread_ && thread_[id]) {
 		assert(!thread_[id]->joinable() && "Shutdown not completed!");
 		thread_[id]->~ParallelHandler();
 		::operator delete(thread_[id]);
-		if (id == 0) {
-			assert(maxThreads_ == 1 && "Shutdown not completed!");
+		thread_[id] = 0;
+		if (id == masterId) {
 			delete [] thread_;
-			thread_     = 0;
-			maxThreads_ = 0;
+			thread_ = 0;
 		}
 	}
 }
 
+inline void ParallelSolve::reportProgress(const SolveEvent& ev) const {
+	return shared_->ctx->reportProgress(ev);
+}
+
 // joins with and destroys all active threads
 void ParallelSolve::joinThreads() {
 	shared_->syncT.reset();
 	shared_->syncT.start();
-	LOG(0, "STARTING  shutdown");
-	int error = thread_[0]->error();
-	for (uint32 i = 1, end = maxThreads_; i != end; ++i) {
-		if (thread_[i]->join() > error) {
-			error = thread_[i]->error();
+	reportProgress(SolveStateEvent(thread_[masterId]->solver(), "shutdown"));
+	error_        = thread_[masterId]->error();
+	uint32 winner = thread_[masterId]->winner() ? uint32(masterId) : UINT32_MAX;
+	for (uint32 i = 1, end = shared_->nextId; i != end; ++i) {
+		if (thread_[i]->join() > error_) {
+			error_ = thread_[i]->error();
+		}
+		if (thread_[i]->winner() && i < winner) {
+			winner = i;
 		}
 		destroyThread(i);
 	}
-	thread_[0]->setError(!shared_->interrupt() ? thread_[0]->error() : error);
-	maxThreads_ = 1;
+	thread_[masterId]->setError(!shared_->interrupt() ? thread_[masterId]->error() : error_);
+	shared_->ctx->setWinner(winner);
+	shared_->nextId = 1;
 	shared_->syncT.stop();
-	LOG(0, "COMPLETED shutdown after %.3fs", shared_->syncT.total());
-}
-
-// attach new client thread
-void ParallelSolve::addThread(Solver& s, SolveParams& p) {
-	uint32 id = shared_->nextId++;
-	assert(id < maxThreads_ && (s.id() == 0 || s.id() == id));
-	assert(thread_ && thread_[id]  && "ParallelSolve::init() not called!");
-	assert(!thread_[id]->joinable()&& "IDs are not unique or thread already initialized!");
-	s.setId(id);
-	thread_[id]->run( EntryPoint(this), s, p ); // start in a new thread
-}
-
-// called by all threads once before solving begins
-// establishes solver<->handler connection and attaches client
-// threads to shared context
-void ParallelSolve::startSolve(Solver& s, const SolveParams& p, const LitVec& assume) {
-	uint32 id = s.id();
-	LOG(id, "ENTERING  algorithm");
-	if (!shared_->terminate()) {
-		if (shared_->ctx->master()->stats.jumps) {
-			s.stats.enableJumpStats();
-		}
-		thread_[id]->attach(s, p);                  // connect solver<->handler
-		shared_->setControl(SharedData::sync_flag); // force initial sync with all threads
-		if (id == 0) {
-			// explicity init parallel handler because Solver::endInit() was
-			// already called.
-			thread_[0]->init(s);
-			shared_->path = &assume;
-			shared_->syncT.start();
-			waitOnSync(s);
-		}
-		else if (!shared_->ctx->attach(s) && !shared_->interrupt()) {
-			bool unsat = !shared_->terminate();
-			if (unsat) { LOG(s.id(), "TRIVIALLY UNSAT!"); }
-			terminate(s, unsat);
-		}
-	}
+	reportProgress(SolveStateEvent(thread_[masterId]->solver(), "shutdown", shared_->syncT.total()));
 }
 
 bool ParallelSolve::initOpt(Solver& s, ValueRep last) {
 	return last != value_true
 	  ||   shared_->enumerator()->continueFromModel(s, false);
 }
-// main solve loop executed by all threads
-bool ParallelSolve::doSolve(Solver& s, const SolveParams& p, const LitVec& assume) {
-	PathPtr a(0); InitParams init = p.init;
-	try {
-		startSolve(s, p, assume);
-		LOG(s.id(), "ENTERING  work loop");
-		for (ValueRep last = value_free; requestWork(s, a);) {
-			thread_[s.id()]->prepareForGP(*a, shared_->maxConflict, a.is_owner());
-			s.stats.reset();
-			if (initOpt(s, last) && initPath(s, *a, init)) {
-				if ((last = solvePath(s, p)) == value_free) {
-					terminate(s, false);
-				}
-				s.clearStopConflict();
-			}
-		}
-	}
-	catch (const std::bad_alloc&)      { exception(s,a,error_oom, "std::bad_alloc" ); }
-	catch (const std::runtime_error& e){ exception(s,a,error_runtime, e.what()); }
-	catch (...)                        { exception(s,a,error_other, "unknown");  }
-	return endSolve(s);
-}
 
-// called once by each solver just after solving process has stopped
-// detaches the solver<->handler connection and joins with
-// all client threads
-bool ParallelSolve::endSolve(Solver& s) {
-	assert(shared_->terminate() || thread_[s.id()]->error() != error_none);
-	LOG(s.id(), "LEAVING   algorithm");
-	uint32 id    = s.id();
-	int    error = thread_[id]->error();
-	thread_[id]->detach();
-	if (!shared_->interrupt() && s.sharedContext() && error == error_none) {
-		s.clearAssumptions();
-		if (s.hasConflict()) {
-			shared_->setControl(SharedData::complete_flag);
-		}
-	}
-	// this thread is leaving
-	shared_->workSem.removeParty(shared_->terminate());
-	s.stats.reset();
-	if (error != error_none) {
-		// solver terminated because of an exception
-		s.reset();
-		s.setId(ParallelSolve::invalidId);
-		std::swap(s.stats.jumps, thread_[id]->aggStats.jumps);
-		std::swap(s.stats.parallel, thread_[id]->aggStats.parallel);
-	}
-	// copy aggregated statistics for this solver to solver
-	s.stats.accu(thread_[id]->aggStats);
-	id == 0
-		? joinThreads()            // join with the other solvers
-		: shared_->ctx->detach(s); // detach client threads from shared context object
-	if (s.stats.parallel) {
-		s.stats.parallel->cpuTime = ThreadTime::getTime();
-	}
-	if (id == 0) {
-		error = thread_[0]->error();
-		if (error != error_none) {
-			switch (error) {
-				case error_oom    : throw std::bad_alloc();
-				case error_runtime: throw std::runtime_error("RUNTIME ERROR!");
-				default:            throw std::runtime_error("UNKNOWN ERROR!");
-			}
-		}
-		shared_->enumerator()->reportResult(shared_->complete());
+// Entry point for master solvers
+bool ParallelSolve::doSolve(Solver& s, const SolveParams& p) {
+	assert(shared_->ctx->master() == &s);
+	// explicity init parallel handler because Solver::endInit() was
+	// already called.
+	allocThread(masterId, s, p);
+	thread_[masterId]->init(s);
+	shared_->path = &getInitialPath();
+	shared_->syncT.start();
+	solveParallel(masterId);
+	joinThreads(); 
+	s.stats.parallel->cpuTime = ThreadTime::getTime();
+	switch(thread_[masterId]->error()) {
+		case error_none   : break;
+		case error_oom    : throw std::bad_alloc();
+		case error_runtime: throw std::runtime_error("RUNTIME ERROR!");
+		default:            throw std::runtime_error("UNKNOWN ERROR!");
 	}
 	return !shared_->complete();
 }
 
-void ParallelSolve::exception(Solver& s, PathPtr& path, ErrorCode e, const char* what) {
+// main solve loop executed by all threads
+void ParallelSolve::solveParallel(uint32 id) {
+	Solver& s           = thread_[id]->solver();
+	const SolveParams p = thread_[id]->params();
+	InitParams init     = p.init;
+	SolveLimits lim     = getSolveLimits();
+	PathPtr a(0);
+	Timer<RealTime> tt; tt.start();
 	try {
-		LOG(s.id(), "EXCEPTION %s", what); (void) what;
-		thread_[s.id()]->setError(e);
-		if (s.id() == 0 || shared_->workSem.active()) { 
+		// establish solver<->handler connection and attach to shared context
+		// should this fail because of an initial conflict, we'll terminate
+		// in requestWork.
+		thread_[id]->attach(*shared_->ctx);
+		reportProgress(SolveStateEvent(s, "algorithm"));
+		for (ValueRep last = value_free; requestWork(s, a);) {
+			thread_[s.id()]->prepareForGP(*a, a.is_owner() ? gp_split : initialGp_, shared_->maxConflict);
+			s.stats.reset();
+			if (initOpt(s, last) && initPath(s, *a, init)) {
+				if ((last = solvePath(s, p, lim)) == value_free) { terminate(s, false); }
+				s.clearStopConflict();
+			}
+		}
+	}
+	catch (const std::bad_alloc&)      { exception(id,a,error_oom, "ERROR: std::bad_alloc" ); }
+	catch (const std::runtime_error& e){ exception(id,a,error_runtime, e.what()); }
+	catch (...)                        { exception(id,a,error_other, "ERROR: unknown");  }
+	assert(shared_->terminate() || thread_[id]->error() != error_none);
+	tt.stop();
+	reportProgress(SolveStateEvent(s, "algorithm", tt.total()));
+	// remove solver<->handler connection and detach from shared context
+	thread_[id]->detach(*shared_->ctx, shared_->interrupt());
+	// this thread is leaving
+	shared_->workSem.removeParty(shared_->terminate());
+}
+
+void ParallelSolve::exception(uint32 id, PathPtr& path, ErrorCode e, const char* what) {
+	try {
+		reportProgress(SolveMsgEvent(thread_[id]->solver(), what));
+		thread_[id]->setError(e);
+		if (id == masterId || shared_->workSem.active()) { 
 			ParallelSolve::terminate();
 			return;
 		}
-		else if (shared_->allowSplit() && path.get()) {
+		else if (path.get() && thread_[id]->disjointPath()) {
 			shared_->workQ.push(path.release());
 			shared_->workSem.up();
 		}
@@ -500,8 +439,7 @@ bool ParallelSolve::requestWork(Solver& s, PathPtr& out) {
 		// only clear path and stop conflict - we don't propagate() here
 		// because we would then have to handle any eventual conflicts
 		if (!s.popRootLevel(s.rootLevel())) {
-			// s has a real top-level conflict - 
-			// there is nothing we can do to continue
+			// s has a real top-level conflict - problem is unsat
 			terminate(s, true);
 		}	
 		else if (shared_->workQ.try_pop(a)) {
@@ -510,7 +448,7 @@ bool ParallelSolve::requestWork(Solver& s, PathPtr& out) {
 			out = a;
 			// do not take over ownership of initial gp!
 			if (a == shared_->path) { out.release(); } 
-			// propagate any new facts now that we have new work
+			// propagate any new facts before starting new work
 			if (s.simplify())       { return true; }
 			// s now has a conflict - either an artifical stop conflict
 			// or a real conflict - we'll handle it in the next iteration
@@ -551,16 +489,14 @@ void ParallelSolve::terminate(Solver& s, bool complete) {
 			// Notify other threads to prepare for solving next priority level.
 			assert(s.decisionLevel() == 0);
 			complete = false;
-			LOG(s.id(), "COMPLETED current optimization step"); 
 			shared_->postMessage(SharedData::msg_sync_unsat, true);
 		}
 		else {
 			shared_->postMessage(SharedData::msg_terminate, true);
-			s.stats.parallel->win = 1;
+			thread_[s.id()]->setWinner();
 			if (complete) { shared_->setControl(SharedData::complete_flag); }
 		}
 	}
-	if (!complete) { s.stats.parallel->term = 1; }
 }
 
 // handles an active synchronization request
@@ -570,7 +506,7 @@ bool ParallelSolve::waitOnSync(Solver& s) {
 		shared_->setControl(SharedData::cancel_restart_flag);
 	}
 	bool hasPath = thread_[s.id()]->hasPath();
-	LOG(s.id(), "WAITING   for synchronization");
+	reportProgress(SolveStateEvent(s, "synchronization"));
 	if (shared_->workSem.wait()) {
 		// last man standing - complete synchronization request
 		shared_->workReq     = 0;
@@ -581,7 +517,7 @@ bool ParallelSolve::waitOnSync(Solver& s) {
 		if (restart) {
 			init = shared_->allowRestart() && !shared_->hasControl(SharedData::cancel_restart_flag);
 			if (init) { shared_->globalR.next(); }
-			shared_->maxConflict = shared_->allowRestart() && shared_->globalR.idx <= maxRestarts_
+			shared_->maxConflict = shared_->allowRestart() && shared_->globalR.idx < maxRestarts_
 				? shared_->globalR.current()
 				: UINT64_MAX;
 		}
@@ -590,12 +526,15 @@ bool ParallelSolve::waitOnSync(Solver& s) {
 			init = false;
 			shared_->setControl(SharedData::terminate_flag | SharedData::complete_flag);
 		}
+		else if (shared_->maxConflict != UINT64_MAX && !shared_->allowRestart()) {
+			shared_->maxConflict = UINT64_MAX;
+		}
 		// add initial path to work queue
 		if (init) { initQueue();  }
 		else      { shared_->setControl(SharedData::restart_abandoned_flag); }
 		shared_->clearControl(SharedData::msg_split | SharedData::msg_sync_unsat | SharedData::msg_sync_restart | SharedData::restart_abandoned_flag | SharedData::cancel_restart_flag);
 		shared_->syncT.stop();
-		LOG(s.id(), "COMPLETED %s after %.3fs", (unsat ? "unsat-sync" : "restart-sync"), shared_->syncT.total());
+		reportProgress(SolveStateEvent(s, (unsat ? "unsat-sync" : "restart-sync"), shared_->syncT.total()));
 		// wake up all blocked threads
 		shared_->workSem.reset();
 	}
@@ -610,14 +549,8 @@ bool ParallelSolve::waitOnSync(Solver& s) {
 //  heuristic for initial splits?
 void ParallelSolve::initQueue() {
 	shared_->clearQueue();
-	int end = numThreads();
-	if (forceGP_) {
-		shared_->setControl(SharedData::allow_gp_flag);
-		end = 1;
-	}
-	else {
-		shared_->setControl(SharedData::forbid_restart_flag);
-	}
+	int end = shared_->hasControl(SharedData::allow_gp_flag) ? 1 : numThreads();
+	assert(end == 1 || shared_->hasControl(SharedData::forbid_restart_flag));
 	for (int i = 0; i != end; ++i) {
 		shared_->workQ.push(shared_->path);
 	}
@@ -638,13 +571,11 @@ bool ParallelSolve::backtrackFromModel(Solver& s) {
 		// grab lock - models must be processed sequentially
 		// in order to simplify printing and to avoid duplicates
 		// in all non-trivial enumeration modes
-		std::unique_lock<tbb::mutex> lock(shared_->modelM);
+		std::lock_guard<std::mutex> lock(shared_->modelM);
 		// first check if the model is still valid once all
 		// information is integrated into the solver
-		uint32 dl = s.decisionLevel();
 		if (shared_->terminate() || !thread_[s.id()]->isModel(s)) {
 			// model no longer a (unique) model or enough models enumerated
-			s.stats.removeModel(dl);
 			return !shared_->terminate(); // continue search?
 		}
 		r = shared_->enumerator()->backtrackFromModel(s, false);
@@ -658,6 +589,8 @@ bool ParallelSolve::backtrackFromModel(Solver& s) {
 		if (shared_->enumerator()->enumerated == 1 && !shared_->enumerator()->supportsRestarts()) {
 			// switch to backtracking based splitting algorithm
 			shared_->setControl(SharedData::forbid_restart_flag | SharedData::allow_gp_flag);
+			// declare the solver's gp as a valid candidate for splitting
+			thread_[s.id()]->setGpType(gp_split); 
 		}
 	}
 	// continue only after we released the mutex to avoid deadlock -
@@ -680,9 +613,8 @@ bool ParallelSolve::handleMessages(Solver& s) {
 	uint32 hId         = s.id();
 	ParallelHandler* h = thread_[hId];
 	if (shared_->terminate()) {
-		LOG(s.id(), "TERMINATE message received!");
+		reportProgress(SolveMsgEvent(s, "TERMINATE message received"));
 		h->handleTerminateMessage();
-		s.stats.parallel->term = 1;
 		s.setStopConflict();
 		return false;
 	}
@@ -693,7 +625,7 @@ bool ParallelSolve::handleMessages(Solver& s) {
 		}
 		return true;
 	}
-	if (s.splittable() && shared_->workReq > 0) {
+	if (h->disjointPath() && s.splittable() && shared_->workReq > 0) {
 		// First declare split request as handled
 		// and only then do the actual split.
 		// This way, we minimize the chance for 
@@ -711,72 +643,83 @@ void ParallelSolve::requestRestart() {
 	}
 }
 
+void ParallelSolveOptions::createSolveObject(SolveAlgorithm*& out, SharedContext& ctx, SolverConfig** sc) const {
+	if (ctx.numSolvers() > 1) {
+		ParallelSolve* x = new ParallelSolve(ctx, *this);
+		out = x;
+		for (uint32 i = 1; i != ctx.numSolvers(); ++i) {
+			x->addSolver(*sc[i]->solver, sc[i]->params);
+		}
+	}
+	else { out = new SimpleSolve(limit); }
+}
 ////////////////////////////////////////////////////////////////////////////////////
 // ParallelHandler
 /////////////////////////////////////////////////////////////////////////////////////////
-ParallelHandler::ParallelHandler(ParallelSolve& ctrl) 
-	: solver_(0)
-	, params_(0)
+ParallelHandler::ParallelHandler(ParallelSolve& ctrl, Solver& s, const SolveParams& p) 
+	: solver_(&s)
+	, params_(&p)
 	, intTail_(0)
 	, error_(0)
+	, win_(0)
 	, messageHandler_(&ctrl) {
+	this->next = this;
 }
 
-ParallelHandler::~ParallelHandler() {
-	detach();
-}
+ParallelHandler::~ParallelHandler() { clearDB(0); }
 
-// adds this as post propagator to s
-void ParallelHandler::attach(Solver& s, const SolveParams& p) {
-	detach();
+// adds this as post propagator to its solver and attaches the solver to ctx.
+bool ParallelHandler::attach(SharedContext& ctx) {
+	assert(solver_ && params_);
 	aggStats.reset();
-	s.stats.enableParallelStats();
-	aggStats.enableParallelStats();
-	if (s.stats.jumps) { aggStats.enableJumpStats(); }
+	aggStats.enableStats(solver_->stats);
 	gp_.reset();
 	gp_.impl= UINT32_MAX;
 	intTail_= 0;
 	error_  = 0;
-	s.addPost(&messageHandler_);
-	s.addPost(this);
-	solver_ = &s;
-	params_ = &p;
+	win_    = 0;
+	up_     = 1;
+	next    = 0;
+	solver_->addPost(&messageHandler_);
+	solver_->addPost(this);
+	return ctx.attach(*solver_);
 }
 
-// detaches this from its solver
-void ParallelHandler::detach() {
-	if (solver_) {
-		handleTerminateMessage();
-		aggStats.accu(solver_->stats);
-		while (!integrated_.empty()) {
-			ClauseHead* c = (ClauseHead*)integrated_.back();
-			integrated_.pop_back();
-			if (error() == 0) {
-				if (c->locked(*solver_)) {
-					solver_->addLearnt(c, c->size());
-				}
-				else {
-					c->destroy(solver_, true);
-				}
-			}
-			else { c->destroy(0, false); }
+// removes this from the list of post propagators of its solver and detaches the solver from ctx.
+void ParallelHandler::detach(SharedContext& ctx, bool fastExit) {
+	handleTerminateMessage();
+	if (solver_->sharedContext() == &ctx) {
+		if (error() == 0 && !fastExit) {
+			clearDB(solver_);
+			solver_->clearAssumptions();
 		}
-		integrated_.clear();
-		intTail_= 0;
-		solver_ = 0;
-		params_ = 0;
+		aggStats.accu(solver_->stats);
+		ctx.detach(*solver_);
+		if (error()) { solver_->reset(); solver_->setId(Solver::invalidId); }
+		solver_->stats.swapStats(aggStats);
+		solver_->stats.parallel->cpuTime = ThreadTime::getTime();
 	}
 }
 
-void ParallelHandler::prepareForGP(const LitVec& out, uint64 restart, bool fromSplit) {
-	gp_.reset(restart, fromSplit);
+void ParallelHandler::clearDB(Solver* s) {
+	for (ClauseDB::iterator it = integrated_.begin(), end = integrated_.end(); it != end; ++it) {
+		ClauseHead* c = static_cast<ClauseHead*>(*it);
+		if (s && c->locked(*s)) { s->addLearnt(c, c->size(), Constraint_t::learnt_other); }
+		else                    { c->destroy(s, s != 0); }
+	}
+	integrated_.clear();
+	intTail_= 0;
+}
+
+void ParallelHandler::prepareForGP(const LitVec& out, GpType t, uint64 restart) {
+	gp_.reset(restart, t);
 	aggStats.parallel->newGP(out.size());
 	aggStats.accu(solver_->stats);
 }
 
 // detach from solver, i.e. ignore any further messages 
 void ParallelHandler::handleTerminateMessage() {
-	if (solver_ && this->next != this) {
+	if (this->next != this) {
 		// mark removed propagators by creating "self-loop"
 		solver_->removePost(&messageHandler_);
 		messageHandler_.next = &messageHandler_;
@@ -794,7 +737,6 @@ void ParallelHandler::handleSplitMessage() {
 	s.pushRootLevel();
 	newGP.push_back(~s.decision(s.rootLevel()));
 	++s.stats.parallel->splits;
-	gp_.split = true;
 	ctrl()->pushWork(newGP);
 }
 
@@ -811,7 +753,7 @@ bool ParallelHandler::simplify(Solver& s, bool sh) {
 		Constraint* c = integrated_[i];
 		if (c->simplify(s, sh)) { 
 			c->destroy(&s, false); 
-			intTail_ -= (i < intTail_); 
+			intTail_ -= (i < intTail_);
 		}
 		else                    { 
 			integrated_[j++] = c;  
@@ -828,12 +770,20 @@ bool ParallelHandler::integrateClauses(Solver& s) {
 	uint32          rec = s.sharedContext()->receive(s, buffer, 30);
 	if (rec != 0) {
 		uint32 intFlags   = ctrl()->integrateFlags();
-		ClauseInfo e; e.setType(Constraint_t::learnt_other);
-		for (uint32 i = 0; i != rec;) {
-			ClauseCreator::Result ret = ClauseCreator::integrate(s, buffer[i++], intFlags, e);
-			if (ret.local) { add(ret.local); }
-			if (!ret.ok)   { while (i != rec) { buffer[i++]->release(); } return false; }				
+		if (s.strategies().updateLbd || params_->reduce.strategy.glue != 0) {
+			intFlags |= ClauseCreator::clause_int_lbd;
 		}
+		ClauseCreator::Result ret; uint32 added = 0;
+		for (uint32 i = 0; i != rec;) {
+			uint32 DL = s.decisionLevel();
+			ret       = ClauseCreator::integrate(s, buffer[i++], intFlags, Constraint_t::learnt_other);
+			added    += ret.status != ClauseCreator::status_subsumed; 
+			if (ret.local)  { add(ret.local); }
+			if (!ret.ok())  { while (i != rec) { buffer[i++]->release(); } break; }
+			if (ret.unit()) { s.stats.addIntegratedAsserting(DL, s.decisionLevel()); }
+		}
+		s.stats.addIntegrated(added);
+		return ret.ok();
 	}
 	return true;
 }
@@ -844,22 +794,26 @@ bool ParallelHandler::propagateFixpoint(Solver& s) {
 	// This is necessary during hierarchical optimization because otherwise
 	// integrating a too strong optimum might irrevocably force the assumption literal
 	// which would defeat its purpose.
-	if ( hasPath() && (s.stats.choices & 63) == 0 && s.isTrue(s.sharedContext()->tagLiteral()) ) {
-		if (!s.sharedContext()->enumerator()->update(s, gp_.split)) {
-			return false;
+	if (up_ == 1 || s.decisionLevel() == s.rootLevel()) {
+		uint32 upMode = s.sharedContext()->updateMode();
+		if (hasPath() && s.isTrue(s.sharedContext()->tagLiteral()) && (upMode == 1 || (s.stats.choices & 63) == 0)) {
+			if (!s.sharedContext()->enumerator()->update(s, disjointPath())) {
+				return false;
+			}
+			if (s.queueSize() != 0 && !s.propagateUntil(this)) {
+				return false;
+			}
 		}
-		if (s.queueSize() != 0 && !s.propagateUntil(this)) {
-			return false;
+		for (;;) {
+			if (!integrateClauses(s))    return false;
+			if (s.queueSize() == 0)      break;
+			if (!s.propagateUntil(this)) return false;
 		}
-	}
-	for (;;) {
-		if (!integrateClauses(s))    return false;
-		if (s.queueSize() == 0)      break;
-		if (!s.propagateUntil(this)) return false;
-	}
-	if (s.stats.conflicts >= gp_.restart) {
-		ctrl()->requestRestart();
-		gp_.restart *= 2;
+		if (s.stats.conflicts >= gp_.restart) {
+			ctrl()->requestRestart();
+			gp_.restart *= 2;
+		}
+		up_ ^= upMode;
 	}
 	return true;
 }
@@ -870,7 +824,7 @@ bool ParallelHandler::isModel(Solver& s) {
 	assert(s.numFreeVars() == 0);
 	// either no unprocessed updates or still a model after
 	// updates were integrated
-	return s.sharedContext()->enumerator()->update(s, gp_.split)
+	return s.sharedContext()->enumerator()->update(s, disjointPath())
 		&& s.numFreeVars() == 0
 		&& s.queueSize()   == 0;
 }
@@ -880,11 +834,12 @@ void ParallelHandler::add(ClauseHead* h) {
 		ClauseHead* o = (ClauseHead*)integrated_[intTail_];
 		integrated_[intTail_] = h;
 		assert(o);
-		if (!ctrl()->integrateUseHeuristic() || o->locked(*solver_) || o->activity().score() > 0) {
-			solver_->addLearnt(o, o->size());
+		if (!ctrl()->integrateUseHeuristic() || o->locked(*solver_) || o->activity().activity() > 0) {
+			solver_->addLearnt(o, o->size(), Constraint_t::learnt_other);
 		}
 		else {
 			o->destroy(solver_, true);
+			solver_->stats.removeIntegrated();
 		}
 	}
 	else {
@@ -898,12 +853,13 @@ void ParallelHandler::add(ClauseHead* h) {
 /////////////////////////////////////////////////////////////////////////////////////////
 // GlobalQueue
 /////////////////////////////////////////////////////////////////////////////////////////
-GlobalQueue::GlobalQueue(uint32 maxT, uint32 types, uint32 lbd) : Distributor(maxT, types, lbd), queue_(0) {
-	assert(maxT < 64);
+GlobalQueue::GlobalQueue(uint32 maxT, uint32 topo) : Distributor(), queue_(0) {
+	assert(maxT < ParallelSolveOptions::supportedSolvers());
 	queue_     = new Queue(maxT);
 	threadId_  = new ThreadInfo[maxT];
 	for (uint32 i = 0; i != maxT; ++i) {
-		threadId_[i].id = queue_->addThread();
+		threadId_[i].id       = queue_->addThread();
+		threadId_[i].peerMask = populatePeerMask(topo, maxT, i);
 	}
 }
 GlobalQueue::~GlobalQueue() {
@@ -913,11 +869,8 @@ void GlobalQueue::release() {
 	if (queue_) {
 		for (uint32 i = 0; i != queue_->maxThreads(); ++i) {
 			Queue::ThreadId& id = getThreadId(i);
-			uint64 mask = uint64(1) << i;
-			for (ClauseNode n; queue_->tryConsume(id, n); ) { 
-				if ((n.targetMask & mask) != 0) {
-					n.lits->release();
-				}
+			for (DistPair n; queue_->tryConsume(id, n); ) { 
+				if (n.sender != i) { n.lits->release(); }
 			}
 		}
 		delete queue_;
@@ -925,23 +878,56 @@ void GlobalQueue::release() {
 		delete [] threadId_;
 	}
 }
-void GlobalQueue::doPublish(const Solver& s, SharedLiterals* lits) {
-	assert(lits->refCount() == queue_->maxThreads());
-	ClauseNode n;
-	n.targetMask = ((uint64(1) << queue_->maxThreads()) - 1) ^ (uint64(1) << s.id());
-	n.lits       = lits;
-	queue_->publish(n, getThreadId(s.id()));
+uint64 GlobalQueue::populateFromCube(uint32 numThreads, uint32 myId, bool ext) const {
+	uint32 n = numThreads;
+	uint32 k = 1;
+	for (uint32 i = n / 2; i > 0; i /= 2, k *= 2) { }
+	uint64 res = 0, x = 1;
+	for (uint32 m = 1; m <= k; m *= 2) {
+		uint32 i = m ^ myId;
+		if      (i < n)         { res |= (x << i);     }
+		else if (ext && k != m) { res |= (x << (i^k)); }
+	}
+	if (ext) {
+		uint32 s = k ^ myId;
+		for(uint32 m = 1; m < k && s >= n; m *= 2) {
+			uint32 i = m ^ s;
+			if (i < n) { res |= (x << i); }
+		}
+	}
+	assert( (res & (x<<myId)) == 0 );
+	return res;
+}
+uint64 GlobalQueue::populatePeerMask(uint32 topo, uint32 maxT, uint32 id) const {
+	switch (topo) {
+		case ParallelSolveOptions::Integration::topo_ring: {
+			uint32 prev = id > 0 ? id - 1 : maxT - 1;
+			uint32 next = (id + 1) % maxT;
+			return Distributor::mask(prev) | Distributor::mask(next);
+		}
+		case ParallelSolveOptions::Integration::topo_cube:  return populateFromCube(maxT, id, false);
+		case ParallelSolveOptions::Integration::topo_cubex: return populateFromCube(maxT, id, true);
+		default:                                     return Distributor::initSet(maxT) ^ Distributor::mask(id);
+	}
+}
+
+void GlobalQueue::publish(const Solver& s, SharedLiterals* n) {
+	assert(n->refCount() >= (queue_->maxThreads()-1));
+	queue_->publish(DistPair(s.id(), n), getThreadId(s.id()));
 }
 uint32 GlobalQueue::receive(const Solver& in, SharedLiterals** out, uint32 maxn) {
 	uint32 r = 0;
 	Queue::ThreadId& id = getThreadId(in.id());
-	uint64 mask = uint64(1) << in.id();
-	for (ClauseNode n; r != maxn && queue_->tryConsume(id, n); ) {
-		if ( (n.targetMask & mask) != 0 ) {
-			out[r++] = n.lits;
+	uint64 peers = getPeerMask(in.id());
+	for (DistPair n; r != maxn && queue_->tryConsume(id, n); ) {
+		if (n.sender != in.id()) {
+			if (inSet(peers, n.sender))  { out[r++] = n.lits; }
+			else if (n.lits->size() == 1){ out[r++] = n.lits; }
+			else                         { n.lits->release(); }
 		}
 	}
 	return r;
 }
 } } // namespace Clasp::mt
+
 #endif
