@@ -556,8 +556,8 @@ bool parseDimacsImpl(std::istream& prg, SharedContext& ctx, bool assertPure, Obj
 	if (p) {
 		ctx.satPrepro.reset(p);
 	}
-	if (assertPure && opt.lits.empty()) {
-		for (int i = 1; ret && i <= maxVar && ret; ++i) {
+	if (assertPure) {
+		for (int i = 1; ret && i <= lastVar && ret; ++i) {
 			uint8 d = (flags[i]>>2);
 			if      (d == 0)  { ret = ctx.addUnary(negLit(i)); }
 			else if (d != 3)  { ret = ctx.addUnary(Literal(i, d != 1)); }
@@ -930,17 +930,18 @@ Input::Format detectFormat(std::istream& in) {
 }
 
 StreamInput::StreamInput(std::istream& in, Format f)
-	: prg_(in), format_(f), maxSat_(false)
+	: prg_(in), format_(f)
 { }
 
-bool StreamInput::read(ApiPtr api, int numModels) {
+bool StreamInput::read(ApiPtr api, uint32 properties) {
 	if (format_ == Input::SMODELS) {
 		return parseLparse(prg_, *api.api);
 	}
 	else if (format_ == Input::DIMACS) {
 		func_.adjust = 0; func_.rhs = -1;
-		bool pure    = abs(numModels) == 1;
-		return parseDimacsImpl(prg_, *api.ctx, pure, func_, maxSat_);
+		bool maxSat  = (properties & AS_MAX_SAT) != 0;
+		bool pure    = (properties & PRESERVE_MODELS) == 0 && (!maxSat || (properties & PRESERVE_MODELS_ON_MIN) == 0);
+		return parseDimacsImpl(prg_, *api.ctx, pure, func_, maxSat);
 	}
 	else {
 		return parseOPB(prg_, *api.ctx, func_);
