@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2010-2011, Benjamin Kaufmann
+// Copyright (c) 2010-2012, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -21,49 +21,31 @@
 #include <clasp/solver.h>
 #include <clasp/clause.h>
 #include <clasp/enumerator.h>
+#if WITH_THREADS
 #include <clasp/util/thread.h>
+#endif
 namespace Clasp {
 /////////////////////////////////////////////////////////////////////////////////////////
 // ShortImplicationsGraph::ImplicationList
 /////////////////////////////////////////////////////////////////////////////////////////
-namespace {
-	inline bool isBinary(const Literal* imp) { return imp->watched(); }
+#if WITH_THREADS
+ShortImplicationsGraph::Block::Block() {
+	for (int i = 0; i != block_cap; ++i) { data[i] = posLit(0); }
+	size_lock = 0;
+	next      = 0;
 }
-
-#ifndef DISABLE_MULTI_THREADING
-struct ShortImplicationsGraph::Block {
-	typedef std::atomic<uint32> atomic_size;
-	enum { block_cap = (64 - (sizeof(atomic_size)+sizeof(SharedBlockPtr)))/sizeof(Literal) };
-	explicit Block() {
-		for (int i = 0; i != block_cap; ++i) { data[i] = posLit(0); }
-		size_lock = 0;
-		next      = 0;
+void ShortImplicationsGraph::Block::addUnlock(uint32 lockedSize, const Literal* x, uint32 xs) {
+	std::copy(x, x+xs, data+lockedSize);
+	size_lock = ((lockedSize+xs) << 1);
+}
+bool ShortImplicationsGraph::Block::tryLock(uint32& size) {
+	uint32 s = size_lock;
+	if ((s & 1) == 0 && size_lock.compare_and_swap(s | 1, s) == s) {
+		size = s >> 1;
+		return true;
 	}
-	const Literal* begin() const { return data; }
-	const Literal* end()   const { return data+size(); }
-	Literal*       end()         { return data+size(); }
-	uint32         size()  const { return size_lock >> 1; }
-	void add(const Literal* x, uint32 a_size, uint32 cs) {
-		std::copy(x, x+a_size, data+cs);
-		if (a_size == 1) {
-			data[cs].watch();
-		}
-	}
-	bool tryLock(uint32& size) {
-		uint32 s = size_lock;
-		if ((s & 1) == 0 && size_lock.compare_and_swap(s | 1, s) == s) {
-			size = s >> 1;
-			return true;
-		}
-		return false;
-	}
-	void unlock(uint32 newSize) {
-		size_lock = (newSize << 1);
-	}
-	SharedBlockPtr next;
-	atomic_size    size_lock;
-	Literal        data[block_cap];
-};
+	return false;
+}
 
 #define FOR_EACH_LEARNT(x, Y) \
 	for (Block* b = (x).learnt; b ; b = b->next) \
@@ -86,20 +68,19 @@ void ShortImplicationsGraph::ImplicationList::clear(bool b) {
 
 void ShortImplicationsGraph::ImplicationList::addLearnt(Literal p, Literal q) {
 	Literal nc[2] = {p, q};
-	uint32 size   = 1 + !isSentinel(q);
+	uint32  ns    = 1 + !isSentinel(q);
+	if (ns == 1) { nc[0].watch(); }
 	for (Block* x;;) {
 		x = learnt;
 		if (x) {
-			uint32 cs;
-			if (x->tryLock(cs)) {
-				if ( (cs + size) <=  Block::block_cap ) {
-					x->add(nc, size, cs);
-					x->unlock(cs + size);
+			uint32 lockedSize;
+			if (x->tryLock(lockedSize)) {
+				if ( (lockedSize + ns) <=  Block::block_cap ) {
+					x->addUnlock(lockedSize, nc, ns);
 				}
 				else {
 					Block* t = new Block();
-					t->add(nc, size, 0);
-					t->unlock(size);
+					t->addUnlock(0, nc, ns);
 					t->next   = x; // x is full and remains locked forever
 					x = learnt= t; // publish new block - unlocks x and learnt
 				}
@@ -123,7 +104,7 @@ bool ShortImplicationsGraph::ImplicationList::hasLearnt(Literal q, Literal r) co
 	FOR_EACH_LEARNT(*this, imp) {
 		if (imp[0] == q || imp[0] == r) {
 			// binary clause subsumes new bin/tern clause
-			if (isBinary(imp))                           { return true; }
+			if (imp->watched())                          { return true; }
 			// existing ternary clause subsumes new tern clause
 			if (!binary && (imp[1] == q || imp[1] == r)) { return true; }
 		}
@@ -137,8 +118,7 @@ void ShortImplicationsGraph::ImplicationList::move(ImplicationList& other) {
 	learnt       = other.learnt;
 	other.learnt = 0;
 }
-#else
-#define FOR_EACH_LEARNT(x, Y) while (const Literal* Y = 0) 
+#undef FOR_EACH_LEARNT
 #endif
 /////////////////////////////////////////////////////////////////////////////////////////
 // ShortImplicationsGraph
@@ -175,7 +155,7 @@ bool ShortImplicationsGraph::addBinary(Literal p, Literal q, bool learnt) {
 		getList(~q).push_left(p);
 		added = 1;
 	}
-#ifndef DISABLE_MULTI_THREADING
+#if WITH_THREADS
 	else if (learnt && !getList(~p).hasLearnt(q)) {
 		getList(~p).addLearnt(q);
 		getList(~q).addLearnt(p);
@@ -196,7 +176,7 @@ bool ShortImplicationsGraph::addTernary(Literal p, Literal q, Literal r, bool le
 		getList(~r).push_right(std::make_pair(p, q));
 		added = 1;
 	}
-#ifndef DISABLE_MULTI_THREADING
+#if WITH_THREADS
 	else if (learnt && !getList(~p).hasLearnt(q, r)) {
 		getList(~p).addLearnt(q, r);
 		getList(~q).addLearnt(p, r);
@@ -248,72 +228,39 @@ void ShortImplicationsGraph::removeTrue(Solver& s, Literal p) {
 	graph_[(~p).index()].clear(true);
 	graph_[ (p).index()].clear(true);
 }
-
-bool ShortImplicationsGraph::propagate(Solver& s, Literal p) const {
-	const ImplicationList& x = graph_[p.index()];
-	if (x.empty()) return true;
-	ImplicationList::const_right_iterator rEnd = x.right_end(); // prefetch
-	// Binary BCP
-	for (ImplicationList::const_left_iterator it = x.left_begin(), end = x.left_end(); it != end; ++it) {
-		if (!s.isTrue(*it) && !s.force(*it, Antecedent(p))) { return false; }
-	}
-	// Ternary BCP
-	Literal q, r;
-	for (ImplicationList::const_right_iterator it = x.right_begin(); it != rEnd; ++it) {
-		if (!s.isTrue(q = it->first) && !s.isTrue(r = it->second)) {
-			if (s.isFalse(q)) { q = r; r = it->first; }
-			if (s.isFalse(r) && !s.force(q, Antecedent(p, ~r))) {
-				return false;
-			}
+struct ShortImplicationsGraph::Propagate {
+	Propagate(Solver& a_s) : s(&a_s) {}
+	bool unary(Literal p, Literal x) const { return s->isTrue(x) || s->force(x, Antecedent(p)); }
+	bool binary(Literal p, Literal x, Literal y) const {
+		ValueRep vx = s->value(x.var()), vy;
+		if (vx != trueValue(x) && (vy=s->value(y.var())) != trueValue(y) && vx + vy) {
+			return vx != 0 ? s->force(y, Antecedent(p, ~x)) : s->force(x, Antecedent(p, ~y));
 		}
+		return true;
 	}
-	// combined short BCP on learnt clauses
-	FOR_EACH_LEARNT(x, imp) {
-		if (!s.isTrue(q = imp[0]) && (isBinary(imp) || !s.isTrue(r = imp[1]))) {
-			if (!isBinary(imp)) {
-				if (s.isFalse(q)) { q = r; r = imp[0]; }
-				if (s.isFalse(r) && !s.force(q, Antecedent(p, ~r))) {
-					return false;
-				}
-			}
-			else if (!s.force(q, Antecedent(p))) {
-				return false;
-			}
-		}
+	Solver* s;
+};
+struct ShortImplicationsGraph::ReverseArc {
+	ReverseArc(const Solver& a_s, uint32 m, Antecedent& o) : s(&a_s), out(&o), maxL(m) {}
+	bool unary(Literal, Literal x) const {
+		if (!isRevLit(*s, x, maxL)) { return true; }
+		*out = Antecedent(~x); 
+		return false;
 	}
-	return true;
-}
-bool ShortImplicationsGraph::reverseArc(const Solver& s, Literal p, uint32 maxLev, Antecedent& out) const {
-	const ImplicationList& x = graph_[p.index()];
-	if (x.empty()) return false;
-	ImplicationList::const_right_iterator rEnd = x.right_end(); // prefetch
-	for (ImplicationList::const_left_iterator it = x.left_begin(), end = x.left_end(); it != end; ++it) {
-		if (isRevLit(s, *it, maxLev)) { 
-			out = Antecedent(~*it);
-			return true;
-		}	
+	bool binary(Literal, Literal x, Literal y) const {
+		if (!isRevLit(*s, x, maxL) || !isRevLit(*s, y, maxL)) { return true; }
+		*out = Antecedent(~x, ~y);
+		return false;
 	}
-	for (ImplicationList::const_right_iterator it = x.right_begin(); it != rEnd; ++it) {
-		if (isRevLit(s, it->first, maxLev) && isRevLit(s, it->second, maxLev)) {
-			out = Antecedent(~it->first, ~it->second);
-			return true;
-		}
-	}
-	FOR_EACH_LEARNT(x, imp) {
-		if (isRevLit(s, imp[0], maxLev) && (isBinary(imp) || isRevLit(s, imp[1], maxLev))) {
-			out = !isBinary(imp) ? Antecedent(~imp[0], ~imp[1]) : Antecedent(~imp[0]);
-			return true;
-		}
-	}
-	return false;
-}
+	const Solver* s; Antecedent* out; uint32 maxL;
+};
+bool ShortImplicationsGraph::propagate(Solver& s, Literal p) const { return forEach(p, Propagate(s)); }
+bool ShortImplicationsGraph::reverseArc(const Solver& s, Literal p, uint32 maxLev, Antecedent& out) const { return !forEach(p, ReverseArc(s, maxLev, out)); }
 bool ShortImplicationsGraph::propagateBin(Assignment& out, Literal p, uint32 level) const {
 	const ImplicationList& x = graph_[p.index()];
 	Antecedent ante(p);
 	for (ImplicationList::const_left_iterator it = x.left_begin(), end = x.left_end(); it != end; ++it) {
-		if (!out.assign(*it, level, p)) {
-			return false;
-		}
+		if (!out.assign(*it, level, p)) { return false; }
 	}
 	return true;
 }
@@ -321,26 +268,28 @@ bool ShortImplicationsGraph::propagateBin(Assignment& out, Literal p, uint32 lev
 // SatPreprocessor
 /////////////////////////////////////////////////////////////////////////////////////////
 SatPreprocessor::~SatPreprocessor() {}
-void SatPreprocessor::reportProgress(char t, uint32 min, uint32 max) {
-	ctx_->enumerator()->reportPreProgress(t, *this, min, max);
-}
-
+void SatPreprocessor::reportProgress(const PreprocessEvent& e) { ctx_->reportProgress(e); }
 /////////////////////////////////////////////////////////////////////////////////////////
 // SharedContext
 /////////////////////////////////////////////////////////////////////////////////////////
-SharedContext::SharedContext() 
-	: symTabPtr_(new SharedSymTab()), master_(new Solver), distributor_(0), lastInit_(0), lastTopLevel_(0)
-	, shareCount_(1), shareConstr_(false), updateShort_(true), frozen_(false) {
+SharedContext::SharedContext(PhysicalSharing x, bool learnImp) 
+	: symTabPtr_(new SharedSymTab()), master_(new Solver), progress_(0), distributor_(0), lastInit_(0), lastTopLevel_(0) {
+	std::memset(&share_, 0, sizeof(Share));
+	share_.count    = 1;
+	share_.physical = x;
+	share_.impl     = learnImp;
 	addVar(Var_t::atom_body_var);
 	problem_.vars = 0;
 	addEnumerator(0);
 	Antecedent::checkPlatformAssumptions();
+	static_assert(sizeof(Share) == sizeof(uint64), "Unsupported Padding");
 }
 
 SharedContext::SharedContext(const SharedContext& rhs,  InitMode) 
-	: master_(new Solver), distributor_(0), lastInit_(0), lastTopLevel_(0), shareCount_(1)
-	, shareConstr_(false), updateShort_(true), frozen_(false) {
-	symTabPtr_ = rhs.symTabPtr_;
+	: master_(new Solver), progress_(0), distributor_(0), lastInit_(0), lastTopLevel_(0)
+	, share_(rhs.share_) {
+	share_.count = 1;
+	symTabPtr_   = rhs.symTabPtr_;
 	++symTabPtr_->refs;
 	addVar(Var_t::atom_body_var);
 	problem_.vars = 0;
@@ -357,6 +306,11 @@ void SharedContext::reset() {
 	new (this) SharedContext();	
 }
 
+void SharedContext::setSolvers(uint32 n) {
+	if (n <= 1) { share_.count = 1; share_.physical = 0; }
+	else        { share_.count = n; }
+}
+
 void SharedContext::reserveVars(uint32 varGuess) {
 	varInfo_.reserve(varGuess);
 }
@@ -395,7 +349,7 @@ void SharedContext::setFrozen(Var v, bool b) {
 	}
 }
 void SharedContext::eliminate(Var v) {
-	assert(validVar(v) && unique() && master()->decisionLevel() == 0); 
+	assert(validVar(v) && !frozen() && master()->decisionLevel() == 0); 
 	if (!eliminated(v)) {
 		varInfo_.toggle(v, VarInfo::ELIM);
 		++problem_.vars_eliminated;
@@ -404,10 +358,9 @@ void SharedContext::eliminate(Var v) {
 	}
 }
 
-
 Solver& SharedContext::startAddConstraints(uint32 constraintGuess) {
-	assert(unique());
-	frozen_ = false;
+	share_.frozen = 0;
+	share_.winner = 0;
 	if (tag_ == negLit(0)) {
 		// add aux var for tag literal
 		tag_ = posLit(addVar(Var_t::atom_var));
@@ -423,73 +376,17 @@ Solver& SharedContext::startAddConstraints(uint32 constraintGuess) {
 	return *master_;
 }
 
-void SharedContext::add(Constraint* c) {
-	assert(unique());
-	++problem_.constraints;
-	master()->constraints_.push_back(c);
-}
-
-bool SharedContext::addUnary(Literal p) {
-	assert(unique());
-	return master()->addUnary(p, Constraint_t::static_constraint);
-}
-
-bool SharedContext::learnUnary(Solver& s, Literal p, ConstraintType t) {
-	assert(t != Constraint_t::static_constraint);
-	s.stats.addLearnt(1, t);
-	if (!unique() && (!updateShortImplications() || btig_.addBinary(p, negLit(0), true))) {
-		distribute(s, p, negLit(0), negLit(0), t);
-	}
-	return true;
-}
-
-bool SharedContext::learnBinary(Solver& s, Literal p, Literal q, const ClauseInfo& info) {
-	assert(info.type() != Constraint_t::static_constraint);
-	bool added = true;
-	if (!updateShortImplications()) {
-		Literal lits[2] = {p, q};
-		s.addLearnt(Clause::newShortClause(s, 0, lits, 2, info), 2);
-	}
-	else if ((added=btig_.addBinary(p, q, true)) == true) {
-		s.stats.addLearnt(2, info.type());
-	}
-	if (added && !unique()) { distribute(s, p, q, negLit(0), info.type()); }
-	return added;
-}
-
-bool SharedContext::learnTernary(Solver& s, Literal p, Literal q, Literal r, const ClauseInfo& info) {
-	assert(info.type() != Constraint_t::static_constraint);
-	bool added = true;
-	if (!updateShortImplications()) {
-		Literal lits[3] = {p, q, r};
-		s.addLearnt(Clause::newShortClause(s, 0, lits, 3, info), 3);
-	}
-	else if ((added=btig_.addTernary(p, q, r, true)) == true) {
-		s.stats.addLearnt(3, info.type());
-	}
-	if (added && !unique()) { distribute(s, p, q, r, info.type()); }
-	return added;
-}
-
-void SharedContext::addPost(PostPropagator* p) {
-	assert(unique());
-	return master()->addPost(p);
-}
-
 void SharedContext::addEnumerator(Enumerator* en) {
-	if (en == 0) {
-		en = new NullEnumerator();
-	}
-	enumerator_.reset(en);
+	enumerator_.reset(en ? en : new NullEnumerator);
 	enumerator_->startInit(*this);
 }
- 
-uint32 SharedContext::numConstraints() const {
-	return numBinary() + numTernary() + master()->constraints_.size();
-}
 
-bool SharedContext::endInit(uint32 shareCount) {
-	assert(unique());
+bool SharedContext::addUnary(Literal p)      { return master()->addUnary(p, Constraint_t::static_constraint); }
+void SharedContext::add(Constraint* c)       { return master()->add(c); }
+uint32 SharedContext::numConstraints() const { return numBinary() + numTernary() + master()->constraints_.size(); }
+
+bool SharedContext::endInit() {
+	assert(!frozen());
 	if (master()->hasConflict()) return false;
 	if (!master()->post_.init(*master())) {
 		return false;
@@ -498,7 +395,7 @@ bool SharedContext::endInit(uint32 shareCount) {
 		~Holder() { if (con) con->destroy(master, true); }
 		Solver*     master;
 		Constraint* con;
-	} enumC = { master(), enumerator()->endInit(*this, shareCount) };
+	} enumC = { master(), enumerator()->endInit(*this, numSolvers()) };
 	if (satPrepro.get()) {
 		SatPrepro temp(satPrepro.release());
 		bool r = temp->preprocess(enumerator()->enumerate());
@@ -509,16 +406,17 @@ bool SharedContext::endInit(uint32 shareCount) {
 	enumC.con = 0;
 	if (!master()->endInit()) return false;
 	lastTopLevel_ = master()->units_;
-	shareCount_   = shareCount;
-	btig_.markShared(shareCount > 1);
+	btig_.markShared(numSolvers() > 1);
+	problem_.constraints        = master()->constraints_.size();
 	problem_.constraints_binary = btig_.numBinary();
 	problem_.constraints_ternary= btig_.numTernary();
-	frozen_ = true;
+	share_.frozen = 1;
 	return true;
 }
 
 bool SharedContext::attach(Solver& other) {
-	assert(!unique() && master_ != &other);
+	assert(frozen());
+	if (other.sharedContext() == this) { return other.popRootLevel(other.rootLevel()); }
 	Var oldV = other.numVars();
 	other.startInit(this, static_cast<uint32>(master_->constraints_.size()-lastInit_));
 	// 1. clone assignment
@@ -557,9 +455,10 @@ bool SharedContext::attach(Solver& other) {
 }
 
 void SharedContext::detach(Solver& s) {
-	s.setEnumerationConstraint(0);
-	if (master() == &s) {
-		shareCount_ = 1;
+	if (s.sharedContext() == this) {
+		s.popRootLevel(s.rootLevel());
+		s.setEnumerationConstraint(0);
+		if (&s != master()) s.shared_ = 0;
 	}
 }
 
@@ -572,18 +471,14 @@ uint32 SharedContext::problemComplexity() const {
 }
 
 SharedLiterals* SharedContext::distribute(const Solver& s, const Literal* lits, uint32 size, const ClauseInfo& extra) const {
-	SharedLiterals* ret = distributor_.get() ? distributor_->publish(s, lits, size, extra) : 0;
-	if (ret && s.stats.parallel) ++s.stats.parallel->shared;
-	return ret;
-}
-
-void SharedContext::distribute(const Solver& s, Literal p, Literal q, Literal r, ConstraintType t) const {
-	if (distributor_.get()) {
-		Literal lits[3] = {p, q, r};
-		SharedLiterals* shared = distributor_->publish(s, lits, 1 + !isSentinel(q) + !isSentinel(r), ClauseInfo().setType(t));
-		if (shared)           { shared->release(); }
-		if (s.stats.parallel) { ++s.stats.parallel->shared; }
+	if (distributor_.get() && !extra.tagged() && (size <= 3 || share_.distribute(size, extra.lbd(), extra.type()))) {
+		uint32 initialRefs = share_.count - (size <= Clause::MAX_SHORT_LEN || !physicalShare(extra.type()));
+		SharedLiterals* x  = SharedLiterals::newShareable(lits, size, extra.type(), initialRefs);
+		distributor_->publish(s, x);
+		const_cast<Solver&>(s).stats.addDistributed(extra.lbd(), extra.type());
+		return initialRefs == share_.count ? x : 0;
 	}
+	return 0;
 }
 
 uint32 SharedContext::receive(const Solver& target, SharedLiterals** out, uint32 maxOut) const {
@@ -595,16 +490,7 @@ uint32 SharedContext::receive(const Solver& target, SharedLiterals** out, uint32
 /////////////////////////////////////////////////////////////////////////////////////////
 // Distributor
 /////////////////////////////////////////////////////////////////////////////////////////
-Distributor::Distributor(uint32 maxShare, uint32 typesToShare, uint32 maxLbd) : maxShare_(maxShare), lbdMax_(maxLbd), typeMask_(typesToShare) {}
+Distributor::Distributor()  {}
 Distributor::~Distributor() {}
 
-SharedLiterals* Distributor::publish(const Solver& source, const Literal* lits, uint32 size, const ClauseInfo& extra) {
-	SharedLiterals* res = 0;
-	if (maxShare_ > 1 && !extra.tagged() && (size <= 3 || (extra.rootLbd() <= lbdMax_ && (extra.type() & typeMask_) != 0))) {
-		res = SharedLiterals::newShareable(lits, size, extra.type(), maxShare_);
-		doPublish(source, res);
-	}
-	return res;
-}
-
 }
