@@ -46,7 +46,7 @@ struct TestingConstraint : public LearntConstraint {
 		LearntConstraint::destroy(s, b);
 	}
 	bool locked(const Solver& s) const { return false; }
-	bool isSatisfied(const Solver&, LitVec&) { return true; }
+	uint32 isOpen(const Solver&, const TypeSet&, LitVec&) { return 0; }
 	static uint32 size() { return uint32(10); }
 	ConstraintType type() const { return type_; }
 	LitVec ante;
@@ -188,12 +188,12 @@ public:
 	void testDefaults() {
 		Solver& s = *ctx.master();
 		const SolverStrategies& x = s.strategies();
-		CPPUNIT_ASSERT(x.heuristic.get() != 0);
+		CPPUNIT_ASSERT(x.heuId == 0);
 		CPPUNIT_ASSERT(x.strRecursive == false);
-		CPPUNIT_ASSERT(x.cflMinAntes == SolverStrategies::all_antes);
+		CPPUNIT_ASSERT(x.ccMinAntes == SolverStrategies::all_antes);
 		CPPUNIT_ASSERT(x.search == SolverStrategies::use_learning);
-		CPPUNIT_ASSERT(x.compress == uint32(250));
-		CPPUNIT_ASSERT(x.randomWatches == false);
+		CPPUNIT_ASSERT(x.compress == 0);
+		CPPUNIT_ASSERT(x.initWatches == SolverStrategies::watch_first);
 
 		CPPUNIT_ASSERT_EQUAL(0u, s.numVars());
 		CPPUNIT_ASSERT_EQUAL(0u, s.numAssignedVars());
@@ -250,7 +250,6 @@ public:
 		Var v2 = ctx.addVar(Var_t::body_var);
 		Solver& s = ctx.startAddConstraints();
 		ctx.eliminate(v1);
-		ctx.endInit();
 		CPPUNIT_ASSERT_EQUAL(uint32(2), s.numVars());
 		CPPUNIT_ASSERT_EQUAL(uint32(1), ctx.numEliminatedVars());
 		CPPUNIT_ASSERT_EQUAL(uint32(1), s.numFreeVars());
@@ -262,6 +261,7 @@ public:
 		// ignore subsequent calls 
 		ctx.eliminate(v1);
 		CPPUNIT_ASSERT_EQUAL(uint32(1), ctx.numEliminatedVars());
+		ctx.endInit();
 	}
 	void testResurrectVar() {
 		/*
@@ -292,8 +292,8 @@ public:
 
 	void testCmpScores() {
 		ReduceStrategy rs;
-		LearntConstraint::Activity a1 = makeActivity(100, 5);
-		LearntConstraint::Activity a2 = makeActivity(90, 3);
+		Activity a1(100, 5);
+		Activity a2(90, 3);
 		CPPUNIT_ASSERT(rs.compare(ReduceStrategy::score_act, a1, a2) > 0);
 		CPPUNIT_ASSERT(rs.compare(ReduceStrategy::score_lbd, a1, a2) < 0);
 		CPPUNIT_ASSERT(rs.compare(ReduceStrategy::score_both, a1, a2) > 0);
@@ -326,7 +326,7 @@ public:
 	}
 
 	void testReset() {
-		ctx.master()->strategies().randomWatches = true;
+		ctx.master()->strategies().initWatches = SolverStrategies::watch_rand;
 		ctx.addVar(Var_t::atom_var); ctx.addVar(Var_t::body_var);
 		Solver& s = ctx.startAddConstraints();
 		s.add( new TestingConstraint(0) );
@@ -636,6 +636,7 @@ public:
 		Literal y = posLit(ctx.addVar(Var_t::atom_var));
 		Literal z = posLit(ctx.addVar(Var_t::atom_var));
 		Solver& s = ctx.startAddConstraints();
+		s.strategies().initWatches = SolverStrategies::watch_least;
 		ClauseCreator cl(&s);
 		cl.start().add(a).add(x).add(y).add(p).end();   // c1
 		cl.start().add(a).add(x).add(y).add(z).end();   // c2
@@ -823,6 +824,7 @@ public:
 		CPPUNIT_ASSERT_EQUAL(false, s.hasConflict());
 		CPPUNIT_ASSERT_EQUAL(true, s.isTrue(posLit(c)));
 		CPPUNIT_ASSERT_EQUAL(0u, s.decisionLevel());
+		CPPUNIT_ASSERT(s.stats.learnts[Constraint_t::learnt_conflict-1] == 1);
 	}
 
 	void testResolveConflict() {
@@ -969,7 +971,7 @@ public:
 			}
 			Literal lit_[2];
 		}*h = new Dummy(negLit(c),negLit(a));
-		ctx.master()->strategies().heuristic.reset(h);
+		ctx.master()->setHeuristic(6, h);
 		Solver& s = ctx.startAddConstraints();
 		ClauseCreator cl(&s);
 		ctx.addBinary(posLit(a), posLit(b));
@@ -1091,7 +1093,7 @@ public:
 					a.push_back( ~ctx->symTab().find(assume[step-1])->lit );
 				}
 			}
-			bool read(ApiPtr p, int) {
+			bool read(ApiPtr p, uint32) {
 				if (step == 0) {
 					p.api->setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "x")
 					    .startRule().addHead(1).addToBody(2, true).endRule()
@@ -1172,11 +1174,11 @@ public:
 		Literal b = posLit(ctx.addVar( Var_t::atom_var ));
 		Literal c = posLit(ctx.addVar( Var_t::atom_var ));
 		Literal d = posLit(ctx.addVar( Var_t::atom_var ));
-		ctx.enableConstraintSharing();
+		ctx.physicalSharing(SharedContext::share_problem);
+		ctx.setSolvers(2);
 		Solver& s = ctx.startAddConstraints();
 		ctx.addBinary(c, d);
-		ctx.endInit(2);
-		SharedContext* other = &ctx;
+		ctx.endInit();
 		s.addBinary(a, b, Constraint_t::learnt_conflict);
 		s.addTernary(~a, ~b, c, Constraint_t::learnt_conflict);
 		CPPUNIT_ASSERT(ctx.numLearntShort() == 2);
@@ -1208,10 +1210,9 @@ public:
 		Literal c = posLit(ctx.addVar( Var_t::atom_var ));
 		Literal d = posLit(ctx.addVar( Var_t::atom_var ));
 		struct Dummy : public Distributor {
-			Dummy() : Distributor(2, 3, 32), unary(0), binary(0), ternary(0) {}
-			void doPublish(const Solver& source, SharedLiterals* lits) {
+			Dummy() : Distributor(), unary(0), binary(0), ternary(0) {}
+			void publish(const Solver& source, SharedLiterals* lits) {
 				uint32 size = lits->size();
-				CPPUNIT_ASSERT(lits->refCount() == 2);
 				unary   += size == 1;
 				binary  += size == 2;
 				ternary += size == 3;
@@ -1230,27 +1231,28 @@ public:
 			uint32 ternary;
 			PodVector<SharedLiterals*>::type shared;
 		}* dummy;
-		ctx.enableLearntSharing(dummy = new Dummy());
+		ctx.setDistribution(UINT32_MAX, UINT32_MAX, UINT32_MAX);
+		ctx.setDistributor(dummy = new Dummy());
+		ctx.setSolvers(2);
 		Solver& s = ctx.startAddConstraints();
-		ctx.endInit(2);
-		SharedContext* other = &ctx;
+		ctx.endInit();
 		LitVec lits; lits.resize(2);
 		lits[0] = a; lits[1] = b;
-		ClauseCreator::create(s, lits, ClauseInfo().setType(Constraint_t::learnt_conflict));
+		ClauseCreator::create(s, lits, ClauseInfo(Constraint_t::learnt_conflict));
 		lits.resize(3);
 		lits[0] = ~a; lits[1] = ~b; lits[2] = ~c;
-		ClauseCreator::create(s, lits, ClauseInfo().setType(Constraint_t::learnt_loop));
+		ClauseCreator::create(s, lits, ClauseInfo(Constraint_t::learnt_loop));
 		lits.resize(1);
 		lits[0] = d;
-		ClauseCreator::create(s, lits, ClauseInfo().setType(Constraint_t::learnt_conflict));
+		ClauseCreator::create(s, lits, ClauseInfo(Constraint_t::learnt_conflict));
 		CPPUNIT_ASSERT(dummy->unary  == 1);
 		CPPUNIT_ASSERT(dummy->binary == 1);
 		CPPUNIT_ASSERT(dummy->ternary == 1);
 		SharedLiterals* rec[3];
 		CPPUNIT_ASSERT(dummy->receive(s, rec, 3) == 3);
-		CPPUNIT_ASSERT(ClauseCreator::integrate(s, rec[0], 0).ok == true);
-		CPPUNIT_ASSERT(ClauseCreator::integrate(s, rec[1], 0).ok == true);
-		CPPUNIT_ASSERT(ClauseCreator::integrate(s, rec[2], 0).ok == true);
+		CPPUNIT_ASSERT(ClauseCreator::integrate(s, rec[0], 0).ok() == true);
+		CPPUNIT_ASSERT(ClauseCreator::integrate(s, rec[1], 0).ok() == true);
+		CPPUNIT_ASSERT(ClauseCreator::integrate(s, rec[2], 0).ok() == true);
 	}
 
 	void testUnfortunateSplitSeq() {
@@ -1258,8 +1260,9 @@ public:
 		Literal b = posLit(ctx.addVar( Var_t::atom_var ));
 		Literal c = posLit(ctx.addVar( Var_t::atom_var ));
 		Literal d = posLit(ctx.addVar( Var_t::atom_var ));
+		ctx.setSolvers(2);
 		Solver& s = ctx.startAddConstraints();
-		ctx.endInit(2);
+		ctx.endInit();
 		
 		Solver s2;
 		ctx.attach(s2);
